from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Rendering.SpatialDistributions.nfw import ProjectedNFW
from pyHalo.Rendering.SpatialDistributions.uniform import LensConeUniform
from pyHalo.concentration_models import preset_concentration_models
from pyHalo.mass_function_models import preset_mass_function_models
from pyHalo.pyhalo import pyHalo
from pyHalo.truncation_models import truncation_models


def WDM(z_lens, z_source, log_mc, sigma_sub=0.025, log_mlow=6., log_mhigh=10., log10_sigma_sub=None,
        mass_function_model_subhalos='LOVELL2020', kwargs_mass_function_subhalos={},
        mass_function_model_fieldhalos='LOVELL2020', kwargs_mass_function_fieldhalos={},
        concentration_model_subhalos='BOSE2016', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='BOSE2016', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3, r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
        mdef_subhalos='TNFW', mdef_field_halos='TNFW', kwargs_density_profile={},
        host_scaling_factor=0.5, redshift_scaling_factor=0.3, two_halo_Lazar_correction=True,
        draw_poisson=True, c_host=None):

    """
    This class generates realizations of dark matter structure in Warm Dark Matter

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param log_mc: the log base 10 of the half-mode mass
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log10_sigma_sub: optional setting of sigma_sub in log10-scale (useful for log-uniform priors); if this is specified
    it overwrites the value of sigma_sub
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param mass_function_model_subhalos: mass function model for subhalos, see mass_function_models.py for a list
    :param kwargs_mass_function_subhalos: keyword arguments for the mass function model
    :param mass_function_model_fieldhalos: mass function model for field halos, see mass_function_models.py for a list
    :param kwargs_mass_function_fieldhalos: keyword arguments for the mass function model
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology (see pyHalo.Cosmology.cosmology)
    :param mdef_subhalos: mass definition for subhalos
    :param mdef_field_halos: mass definition for field halos
    :param kwargs_density_profile: keyword arguments for the specified mass profile
    :param host_scaling_factor: the scaling with host halo mass of the projected number density of subhalos
    :param redshift_scaling_factor: the scaling with (1+z) of the projected number density of subhalos
    :param two_halo_Lazar_correction: bool; if True, adds the correction to the two-halo contribution from around the
    main deflector presented by Lazar et al. (2021)
    :param c_host: manually set host halo concentration
    :return: a realization of dark matter halos
    """
    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)
    # WE ALSO SPECIFY THE GEOMETRY OF THE RENDERING VOLUME
    geometry = Geometry(pyhalo.cosmology, z_lens, z_source,
                        cone_opening_angle_arcsec, geometry_type)

    # SET THE SPATIAL DISTRIBUTION MODELS FOR SUBHALOS AND FIELD HALOS:
    subhalo_spatial_distribution = ProjectedNFW
    fieldhalo_spatial_distribution = LensConeUniform

    # SET THE MASS FUNCTION MODELS FOR SUBHALOS AND FIELD HALOS
    # NOTE: MASS FUNCTIONS SHOULD NOT BE INSTANTIATED HERE
    mass_function_model_subhalos, kwargs_mass_function_subhalos = preset_mass_function_models(mass_function_model_subhalos,
                                                                                              kwargs_mass_function_subhalos)
    kwargs_mass_function_subhalos['log_mc'] = log_mc

    mass_function_model_fieldhalos, kwargs_mass_function_fieldhalos = preset_mass_function_models(mass_function_model_fieldhalos,
                                                                                                  kwargs_mass_function_fieldhalos)
    kwargs_mass_function_fieldhalos['log_mc'] = log_mc

    # SET THE CONCENTRATION-MASS RELATION FOR SUBHALOS AND FIELD HALOS
    kwargs_concentration_model_subhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_subhalos['log_mc'] = log_mc
    kwargs_concentration_model_fieldhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_fieldhalos['log_mc'] = log_mc

    model_subhalos, kwargs_mc_subs = preset_concentration_models(concentration_model_subhalos,
                                                                 kwargs_concentration_model_subhalos)
    concentration_model_CDM = preset_concentration_models('DIEMERJOYCE19')[0]
    kwargs_mc_subs['concentration_cdm_class'] = concentration_model_CDM
    concentration_model_subhalos = model_subhalos(**kwargs_mc_subs)

    model_fieldhalos, kwargs_mc_field = preset_concentration_models(concentration_model_fieldhalos,
                                                                    kwargs_concentration_model_fieldhalos)
    kwargs_mc_field['cosmo'] = pyhalo.astropy_cosmo
    kwargs_mc_field['log_mc'] = log_mc
    concentration_model_CDM = preset_concentration_models('DIEMERJOYCE19')[0]
    kwargs_mc_field['concentration_cdm_class'] = concentration_model_CDM
    concentration_model_fieldhalos = model_fieldhalos(**kwargs_mc_field)
    if c_host is None:
        c_host = concentration_model_fieldhalos.nfw_concentration(10 ** log_m_host, z_lens)

    # SET THE TRUNCATION RADIUS FOR SUBHALOS AND FIELD HALOS
    kwargs_truncation_model_subhalos['lens_cosmo'] = pyhalo.lens_cosmo
    kwargs_truncation_model_fieldhalos['lens_cosmo'] = pyhalo.lens_cosmo
    model_subhalos, kwargs_trunc_subs = truncation_models(truncation_model_subhalos)
    kwargs_trunc_subs.update(kwargs_truncation_model_subhalos)
    if truncation_model_subhalos in ['TRUNCATION_GALACTICUS_KEELEY24', 'TRUNCATION_GALACTICUS']:
        kwargs_trunc_subs['c_host'] = c_host
    kwargs_trunc_subs['lens_cosmo'] = pyhalo.lens_cosmo
    truncation_model_subhalos = model_subhalos(**kwargs_trunc_subs)

    model_fieldhalos, kwargs_trunc_field = truncation_models(truncation_model_fieldhalos)
    kwargs_trunc_field.update(kwargs_truncation_model_fieldhalos)
    kwargs_trunc_field['lens_cosmo'] = pyhalo.lens_cosmo
    truncation_model_fieldhalos = model_fieldhalos(**kwargs_trunc_field)

    # NOW THAT THE CLASSES ARE SPECIFIED, WE SORT THE KEYWORD ARGUMENTS AND CLASSES INTO LISTS
    population_model_list = ['SUBHALOS', 'LINE_OF_SIGHT', 'TWO_HALO']
    # check for log10 value of sigma_sub
    if log10_sigma_sub is not None:
        sigma_sub = 10 ** log10_sigma_sub
    mass_function_class_list = [mass_function_model_subhalos,
                                mass_function_model_fieldhalos,
                                mass_function_model_fieldhalos]
    kwargs_mass_function_subhalos.update({'log_mlow': log_mlow,
                       'log_mhigh': log_mhigh,
                       'm_pivot': 10 ** 8,
                       'power_law_index': shmf_log_slope,
                       'log_m_host': log_m_host,
                       'sigma_sub': sigma_sub,
                       'delta_power_law_index': 0.0,
                       'host_scaling_factor': host_scaling_factor,
                       'redshift_scaling_factor': redshift_scaling_factor,
                                          'draw_poisson': draw_poisson})
    kwargs_mass_function_fieldhalos.update({'log_mlow': log_mlow,
                  'log_mhigh': log_mhigh,
                  'LOS_normalization': LOS_normalization,
                  'm_pivot': 10 ** 8,
                  'log_m_host': log_m_host,
                  'delta_power_law_index': 0.0, 'draw_poisson': draw_poisson})
    kwargs_mass_function_list = [kwargs_mass_function_subhalos, kwargs_mass_function_fieldhalos, kwargs_mass_function_fieldhalos]
    spatial_distribution_class_list = [subhalo_spatial_distribution, fieldhalo_spatial_distribution, fieldhalo_spatial_distribution]

    kwargs_subhalos_spatial = {'m_host': 10 ** log_m_host, 'zlens': z_lens,
                               'rmax2d_arcsec': cone_opening_angle_arcsec / 2, 'r_core_units_rs': r_tidal,
                               'lens_cosmo': pyhalo.lens_cosmo, 'c_host': c_host}
    kwargs_los_spatial = {'cone_opening_angle': cone_opening_angle_arcsec, 'geometry': geometry}
    kwargs_spatial_distribution_list = [kwargs_subhalos_spatial, kwargs_los_spatial, kwargs_los_spatial]
    kwargs_halo_model = {'truncation_model_subhalos': truncation_model_subhalos,
                         'concentration_model_subhalos': concentration_model_subhalos,
                         'truncation_model_field_halos': truncation_model_fieldhalos,
                         'concentration_model_field_halos': concentration_model_fieldhalos,
                         'kwargs_density_profile': kwargs_density_profile}
    realization_list = pyhalo.render(population_model_list, mass_function_class_list, kwargs_mass_function_list,
                                     spatial_distribution_class_list, kwargs_spatial_distribution_list,
                                     geometry, mdef_subhalos, mdef_field_halos, kwargs_halo_model,
                                     two_halo_Lazar_correction,
                                     nrealizations=1)
    return realization_list[0]


def WDM_mixed(z_lens, z_source, log_mc, mixed_DM_frac, sigma_sub=0.025, log_mlow=6., log_mhigh=10.,
        mass_function_model_subhalos='SHMF_MIXED_WDM_TURNOVER', kwargs_mass_function_subhalos={},
        mass_function_model_fieldhalos='MIXED_WDM_TURNOVER', kwargs_mass_function_fieldhalos={},
        concentration_model_subhalos='BOSE2016', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='BOSE2016', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3, r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
        kwargs_density_profile={}):

    """
    Implements the mixed dark matter model presented by Keely et al. (2023)
    https://arxiv.org/pdf/2301.07265.pdf

    :param z_lens:
    :param z_source:
    :param log_mc:
    :param mixed_DM_frac: fraction of dark matter in CDM component
    :param sigma_sub:
    :param log_mlow:
    :param log_mhigh:
    :param mass_function_model_subhalos:
    :param kwargs_mass_function_subhalos:
    :param mass_function_model_fieldhalos:
    :param kwargs_mass_function_fieldhalos:
    :param concentration_model_subhalos:
    :param kwargs_concentration_model_subhalos:
    :param concentration_model_fieldhalos:
    :param kwargs_concentration_model_fieldhalos:
    :param truncation_model_subhalos:
    :param kwargs_truncation_model_subhalos:
    :param truncation_model_fieldhalos:
    :param kwargs_truncation_model_fieldhalos:
    :param shmf_log_slope:
    :param cone_opening_angle_arcsec:
    :param log_m_host:
    :param r_tidal:
    :param LOS_normalization:
    :param geometry_type:
    :param kwargs_cosmo:
    :param kwargs_density_profile:
    :return:
    """
    params = ['a_wdm', 'b_wdm', 'c_wdm']
    values_keeley_2023 = [0.5, 0.8, -3.0]
    for i, param in enumerate(params):
        if param not in kwargs_mass_function_subhalos.keys():
            kwargs_mass_function_subhalos[param] = values_keeley_2023[i]
        if param not in kwargs_mass_function_fieldhalos.keys():
            kwargs_mass_function_fieldhalos[param] = values_keeley_2023[i]
    kwargs_mass_function_subhalos['mixed_DM_frac'] = mixed_DM_frac
    kwargs_mass_function_fieldhalos['mixed_DM_frac'] = mixed_DM_frac
    kwargs_wdm = {'z_lens': z_lens, 'z_source': z_source, 'log_mc': log_mc, 'sigma_sub': sigma_sub,
                  'log_mlow': log_mlow, 'log_mhigh': log_mhigh,
                  'mass_function_model_subhalos': mass_function_model_subhalos,
                  'kwargs_mass_function_subhalos': kwargs_mass_function_subhalos,
                  'mass_function_model_fieldhalos': mass_function_model_fieldhalos,
                  'kwargs_mass_function_fieldhalos': kwargs_mass_function_fieldhalos,
                  'concentration_model_subhalos': concentration_model_subhalos,
                  'kwargs_concentration_model_subhalos': kwargs_concentration_model_subhalos,
                  'concentration_model_fieldhalos': concentration_model_fieldhalos,
                  'kwargs_concentration_model_fieldhalos': kwargs_concentration_model_fieldhalos,
                  'truncation_model_subhalos': truncation_model_subhalos,
                  'kwargs_truncation_model_subhalos': kwargs_truncation_model_subhalos,
                  'truncation_model_fieldhalos': truncation_model_fieldhalos,
                  'kwargs_truncation_model_fieldhalos': kwargs_truncation_model_fieldhalos,
                  'shmf_log_slope': shmf_log_slope, 'cone_opening_angle_arcsec': cone_opening_angle_arcsec,
                  'log_m_host': log_m_host, 'r_tidal': r_tidal, 'LOS_normalization': LOS_normalization,
                  'geometry_type': geometry_type, 'kwargs_cosmo': kwargs_cosmo,
                  'kwargs_density_profile': kwargs_density_profile
                  }
    return WDM(**kwargs_wdm)


def WDMGeneral(z_lens, z_source, log_mc, dlogT_dlogk, sigma_sub=0.025, log_mlow=6., log_mhigh=10., log10_sigma_sub=None,
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3, r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
        mdef_subhalos='TNFW', mdef_field_halos='TNFW', kwargs_density_profile={},
        host_scaling_factor=0.5, redshift_scaling_factor=0.3, two_halo_Lazar_correction=True, c_host=None):

    """
    This preset model implements a generalized treatment of warm dark matter, or any theory that produces a cutoff in
    the linear matter power spectrum. One additional free parameter is added to parameterize the cutoff in the matter
     power spectrum, dlogT_dlogk, which is the absolute value of the logarithmic derivative of the transfer function
     evaluated at k_1/2, the wavenumber corresponding to the half-mode mass (or log_mc).

     In this model, halo concentrations are computed following the formalism suggeesteed by Schneider et al. (2012) and
     Ludlow et al. (2016), in which the central density of a halo more or less is fully determined by the formation time.

     The mass function is computed from dlogT_dlogk based on the prescription by Stucker et al. (2021)
     https://arxiv.org/pdf/2109.09760.pdf

    :param z_lens: the lens redshift
    :param z_source: source redshift
    :param log_mc: the log (base 10) of the half-mode mass
    :param dlogT_dlogk: the absolute value of the logarithmic derivative of the transfer function at k_1/2; the model
    is calibrated for values between ~1 and ~3
    :param sigma_sub: amplitude of the subhalo mass function
    :param log_mlow: minimum halo mass to render
    :param log_mhigh: maximum halo mass to render
    :param log10_sigma_sub: optional keyword argument that overrides sigma_sub; specified in a log10-scale
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: logarithmic slope of the subhalo mass function
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume
    :param log_m_host: the log (base 10) of the host halo mass
    :param r_tidal: the core size in units of the scale radius of the host halo; subhalos are rendered uniformly in 3D
    inside this radius
    :param LOS_normalization: the amplitude of the LOS mass function relative to Sheth-Tormen
    :param geometry_type: CONE, DOUBLE_CONE, CYLINDER - sets the geometry of the rendering volume
    :param kwargs_cosmo: keyword arguments to set cosmology
    :param mdef_subhalos: mass definition for subhalos
    :param mdef_field_halos: mass definition for field halos
    :param kwargs_density_profile: keyword arguments for the density profile
    :param host_scaling_factor: the scaling with host halo mass of the projected number density of subhalos
    :param redshift_scaling_factor: the scaling with (1+z) of the projected number density of subhalos
    :param two_halo_Lazar_correction: bool; if True, adds the correction to the two-halo contribution from around the
    main deflector presented by Lazar et al. (2021)
    :param c_host: manually fix the host halo concentration
    :return:
    """
    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)
    # WE ALSO SPECIFY THE GEOMETRY OF THE RENDERING VOLUME
    geometry = Geometry(pyhalo.cosmology, z_lens, z_source,
                        cone_opening_angle_arcsec, geometry_type)

    # SET THE SPATIAL DISTRIBUTION MODELS FOR SUBHALOS AND FIELD HALOS:
    subhalo_spatial_distribution = ProjectedNFW
    fieldhalo_spatial_distribution = LensConeUniform
    kwargs_model_dlogT_dlogk = {'dlogT_dlogk': dlogT_dlogk}
    mass_function_model_subhalos, kwargs_mfunc_subs = preset_mass_function_models('STUCKER_SHMF', kwargs_model_dlogT_dlogk)
    mass_function_model_fieldhalos, kwargs_mfunc_field = preset_mass_function_models('STUCKER', kwargs_model_dlogT_dlogk)
    # SET THE CONCENTRATION-MASS RELATION FOR SUBHALOS AND FIELD HALOS
    concentration_model = 'LUDLOW_WDM'
    model_subhalos, kwargs_concentration_model_subhalos = preset_concentration_models(concentration_model,
                                                                                      kwargs_model_dlogT_dlogk)

    kwargs_concentration_model_subhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_subhalos['log_mc'] = log_mc
    concentration_model_subhalos = model_subhalos(**kwargs_concentration_model_subhalos)

    model_fieldhalos, kwargs_concentration_model_fieldhalos = preset_concentration_models(concentration_model,
                                                                                          kwargs_model_dlogT_dlogk)
    kwargs_concentration_model_fieldhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_fieldhalos['log_mc'] = log_mc

    concentration_model_fieldhalos = model_fieldhalos(**kwargs_concentration_model_fieldhalos)
    concentration_model_CDM = preset_concentration_models('DIEMERJOYCE19')[0](pyhalo.astropy_cosmo)
    if c_host is None:
        c_host = concentration_model_CDM.nfw_concentration(10 ** log_m_host, z_lens)
    # SET THE TRUNCATION RADIUS FOR SUBHALOS AND FIELD HALOS
    kwargs_truncation_model_subhalos['lens_cosmo'] = pyhalo.lens_cosmo
    kwargs_truncation_model_fieldhalos['lens_cosmo'] = pyhalo.lens_cosmo
    model_subhalos, kwargs_trunc_subs = truncation_models(truncation_model_subhalos)
    kwargs_trunc_subs.update(kwargs_truncation_model_subhalos)
    if truncation_model_subhalos in ['TRUNCATION_GALACTICUS_KEELEY24', 'TRUNCATION_GALACTICUS']:
        kwargs_trunc_subs['c_host'] = c_host
    truncation_model_subhalos = model_subhalos(**kwargs_trunc_subs)
    model_fieldhalos, kwargs_trunc_field = truncation_models(truncation_model_fieldhalos)
    kwargs_trunc_field.update(kwargs_truncation_model_fieldhalos)
    truncation_model_fieldhalos = model_fieldhalos(**kwargs_trunc_field)
    # NOW THAT THE CLASSES ARE SPECIFIED, WE SORT THE KEYWORD ARGUMENTS AND CLASSES INTO LISTS
    population_model_list = ['SUBHALOS', 'LINE_OF_SIGHT', 'TWO_HALO']
    mass_function_class_list = [mass_function_model_subhalos,
                                mass_function_model_fieldhalos,
                                mass_function_model_fieldhalos]
    # check for log10 value of sigma_sub
    if log10_sigma_sub is not None:
        sigma_sub = 10 ** log10_sigma_sub
    kwargs_mfunc_subs.update({'log_mlow': log_mlow,
                       'log_mhigh': log_mhigh,
                       'm_pivot': 10 ** 8,
                       'power_law_index': shmf_log_slope,
                       'log_m_host': log_m_host,
                       'sigma_sub': sigma_sub,
                       'delta_power_law_index': 0.0,
                        'log_mc': log_mc,
                        'host_scaling_factor': host_scaling_factor,
                        'redshift_scaling_factor': redshift_scaling_factor})
    kwargs_mfunc_field.update({'log_mlow': log_mlow,
                  'log_mhigh': log_mhigh,
                  'LOS_normalization': LOS_normalization,
                  'm_pivot': 10 ** 8,
                  'log_m_host': log_m_host,
                  'delta_power_law_index': 0.0,
                  'log_mc': log_mc})
    kwargs_mass_function_list = [kwargs_mfunc_subs, kwargs_mfunc_field, kwargs_mfunc_field]
    spatial_distribution_class_list = [subhalo_spatial_distribution, fieldhalo_spatial_distribution, fieldhalo_spatial_distribution]
    kwargs_subhalos_spatial = {'m_host': 10 ** log_m_host, 'zlens': z_lens,
                               'rmax2d_arcsec': cone_opening_angle_arcsec / 2, 'r_core_units_rs': r_tidal,
                               'lens_cosmo': pyhalo.lens_cosmo, 'c_host': c_host}
    kwargs_los_spatial = {'cone_opening_angle': cone_opening_angle_arcsec, 'geometry': geometry}
    kwargs_spatial_distribution_list = [kwargs_subhalos_spatial, kwargs_los_spatial, kwargs_los_spatial]

    kwargs_halo_model = {'truncation_model_subhalos': truncation_model_subhalos,
                         'concentration_model_subhalos': concentration_model_subhalos,
                         'truncation_model_field_halos': truncation_model_fieldhalos,
                         'concentration_model_field_halos': concentration_model_fieldhalos,
                         'kwargs_density_profile': kwargs_density_profile}

    realization_list = pyhalo.render(population_model_list, mass_function_class_list, kwargs_mass_function_list,
                                     spatial_distribution_class_list, kwargs_spatial_distribution_list,
                                     geometry, mdef_subhalos, mdef_field_halos, kwargs_halo_model,
                                     two_halo_Lazar_correction,
                                     nrealizations=1)

    return realization_list[0]



import numpy.testing as npt
import pytest
from pyHalo.concentration_models import preset_concentration_models
import numpy as np

class _CustomMC(object):

    def __init__(self, a, b):

        self._a = a
        self._b = b

    def nfw_concentration(self, m, z):
        return self._a * (m/10**8) ** self._b * (1 + z) ** -1

class _CustomMCInvalid(object):

    def return_concentration(self, m, z):
        return 16.0 * (m / 10 ** 8) ** -0.2 * (1 + z) ** -0.5

class TestConcentrationModels(object):

    def setup_method(self):

        self.model_list = ['DIEMERJOYCE19', 'PEAK_HEIGHT_POWERLAW', 'WDM_HYPERBOLIC',
                           'WDM_POLYNOMIAL', 'BOSE2016', 'LAROCHE2022', 'LUDLOW2016', 'LUDLOW_WDM']
        self.model_names = ['DIEMERJOYCE19', 'PEAK_HEIGHT_POWERLAW', 'WDM_HYPERBOLIC',
                            'WDM_POLYNOMIAL', 'WDM_POLYNOMIAL', 'WDM_POLYNOMIAL', 'LUDLOW2016', 'LUDLOW_WDM']

    def test_models(self):

        for model, model_name in zip(self.model_list, self.model_names):
            mod, kw = preset_concentration_models(model)
            npt.assert_string_equal(model_name, mod.name)

    def test_custom_class(self):

        kwargs_model = {'custom_class': _CustomMC, 'a': 10.0, 'b': -1.0}
        mod, kw = preset_concentration_models('CUSTOM', kwargs_model)
        concentration_model = mod(**kw)
        c = concentration_model.nfw_concentration(10 ** 7.5, 0.5)
        npt.assert_almost_equal(c, kwargs_model['a'] * (10 ** 7.5 / 10 ** 8) ** kwargs_model['b'] / 1.5)

        kwargs_model = {'custom_class': _CustomMCInvalid}
        args = ('CUSTOM', kwargs_model)
        npt.assert_raises(Exception, preset_concentration_models, args)

    def test_formation_history(self):

        kwargs_model = {'dlogT_dlogk': 2.0}
        mod, kw = preset_concentration_models('FROM_FORMATION_HISTORY', kwargs_model)

if __name__ == '__main__':
     pytest.main()



from abc import ABC, abstractmethod
import numpy as np
from pyHalo.utilities import ITSampling
from scipy.integrate import quad

_log10_rperi_bins = np.array([-2.   , -1.885, -1.77 , -1.655, -1.54 , -1.425, -1.31 , -1.195,
       -1.08 , -0.965, -0.85 , -0.735, -0.62 , -0.505, -0.39 , -0.275,
       -0.16 , -0.045,  0.07 ,  0.185])
_log10_prob = np.array([  18,   23,   37,   33,   42,   78,  139,  181,  321,  538,  749,
       1018, 1268, 1501, 1577, 1232,  569,  207,   23,    7])
_cdf = np.cumsum(_log10_prob)
_log10_rpericenter_sampling = ITSampling(_log10_rperi_bins, _cdf)

class Halo(ABC):

    def __init__(self, mass=None, x=None, y=None, r3d=None, mdef=None, z=None,
                 sub_flag=None, lens_cosmo_instance=None, args={}, unique_tag=None, fixed_position=False):

        """
        This is the main class for objects rendered in the lens volume. It keeps track of stuff like the position,
        mass, redshift, and structural properties (e.g. concentration, core radius, etc.)

        :param mass: halo mass in M_sun
        :param x: angular coordinate x in arcsec
        :param y: angular coordinate y in arcsec
        :param r3d: 3D position of halo in kpc (used to compute the truncation radius for subhalos)
        For field halos this is not relevant and is set to None
        :param mdef: mass definition for the halo
        :param z: halo redshift
        :param sub_flag: bool; if True, the halo is treated as a main deflector subhalo
        :param lens_cosmo_instance: an instance of LensCosmo
        :param args: keyword arguments that include default settings for the halo
        :param unique_tag: a random number with 16 decimal places that uniquely identifies each halo
        :param fixed_position: determiens whether halos can be moved around when aligning a realization with
        the rendering volume
        """

        self.lens_cosmo = lens_cosmo_instance
        self.mass = mass
        # x and y in arcsec
        self.x = x
        self.y = y
        self.r3d = r3d
        self.mdef = mdef
        self.z = z
        self.is_subhalo = sub_flag
        self._args = args
        self.unique_tag = unique_tag
        self._rescale_norm = 1.
        self.fixed_position = fixed_position
        self._rescaled_once = False
        assert z > 0, 'Negative redshifts are unphysical for halos'

    def rescale_normalization(self, factor):
        """
        Sets the rescaling factor for the normalization (only can do this once)
        :param factor:
        :return:
        """
        if self._rescaled_once:
            pass
        else:
            self._rescaled_once = True
            self._rescale_norm = factor
            if hasattr(self, '_params_physical'):
                delattr(self, '_params_physical')
            if hasattr(self, '_kwargs_lenstronomy'):
                delattr(self, '_kwargs_lenstronomy')
            if hasattr(self, '_nfw_params'):
                delattr(self, '_nfw_params')

    @property
    @abstractmethod
    def profile_args(self):
        """
        This routine computes properties of the halo required to specify it
        """
        ...

    @property
    def params_physical(self):
        raise Exception('this halo class does not have attribute params_physical')

    @property
    @abstractmethod
    def lenstronomy_ID(self):
        """
        Returns a list of profile names recognized by lenstronomy

        Example:
            a truncated NFW profile would return ['TNFW']
            a hybrid NFW and point mass profile would return ['NFW', 'POINT_MASS']

        """
        ...

    @property
    @abstractmethod
    def lenstronomy_params(self):
        """
        Returns a list of keyword arguments for the profile, must be the same length as lenstronomy_ID
        :return:
        """
        ...

    @property
    def z_infall(self):

        """
        Evaluate the infall redshift using a PDF generated by galacticus.
        Note: This routine is meaningless and therefore not used for LOS halos
        :return: the infall redshift of a halo assuming it is in a host halo at redshift self.z
        """

        if not hasattr(self, '_z_infall'):
            self._z_infall = self.lens_cosmo.z_accreted_from_zlens(self.mass, self.z)
            assert self._z_infall >= self.z, 'infall redshifts less than current redshift are unphysical'
        return self._z_infall

    @property
    def time_since_infall(self):
        """
        This routine calculates the time in Gyr since infall for subhalos using the infall redshift as predicted by
        Galacticus
        :return: the time since the subhalo was accreted onto thehost [Gyr]
        """
        if not self.is_subhalo:
            print("time since infall is a meaningless concept for field halos")
            return None
        if not hasattr(self, '_time_since_infall'):
            astropy = self.lens_cosmo.cosmo.astropy
            self._time_since_infall = astropy.age(self.z).value - astropy.age(self.z_infall).value
            assert self._time_since_infall > 0
        return self._time_since_infall

    @property
    def rperi_units_r200(self):
        """
        Returns the orbital pericenter of a subhalo in units of the host halo virial radius. This method
        uses output from the semi-analytic model Galacticus
        :return:
        """
        if not self.is_subhalo:
            print("Orbital pericenter is a meaningless concept for field halos. It is possible you assigned a tidal "
                  "truncation model that requires this information to field halos.")
            return None
        if not hasattr(self, '_rperi_units_r200'):
            self._rperi_units_r200 = 10**float(_log10_rpericenter_sampling(n_samples=1.0))
        return self._rperi_units_r200

    @property
    def bound_mass(self):
        raise Exception('this halo class does not have a bound mass attribute because the profile does not have '
                        'a tidal truncation radius')

    @property
    def halo_age(self):
        """Computes the age of a halo assuming it collapsed at z=10
        TODO: Make this a function of halo mass
        """
        if not hasattr(self, '_halo_age'):
            self._halo_age = self.lens_cosmo.cosmo.halo_age(self.z, zform=10)
        return self._halo_age

    def density_profile_3d(self, *args, **kwargs):
        raise Exception('This profile does not have a density_profile_3d method defined in the class')

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        raise Exception('this class does not have a well-defined concentration parameter')

    @property
    def z_eval(self):
        """
        Returns the redshift at which to evaluate the concentration-mass relation
        """
        if not hasattr(self, '_zeval'):

            if self.is_subhalo:
                if 'evaluate_mc_at_zlens' in self._args.keys() and self._args['evaluate_mc_at_zlens']:
                    self._zeval = self.z
                else:
                    self._zeval = self.z_infall
            else:
                self._zeval = self.z

        return self._zeval

    @property
    def nfw_params(self):
        """
        Computes the nfw profile parameters (rs,rho_s) from mass and concentration
        :return: rs, r200 and rho_s in units kpc, kpc, and M_sun / kpc^3
        """
        if not hasattr(self, '_nfw_params'):
            if self.mdef in ['TNFWC', 'GNFW']:
                pseudo_nfw = True
            else:
                pseudo_nfw = False
            rhos, rs, r200 = self.lens_cosmo.NFW_params_physical(self.mass, self.c, self.z, pseudo_nfw)
            self._nfw_params = [rhos, rs, r200]
        return self._nfw_params[0], self._nfw_params[1], self._nfw_params[2]

    def mass_3d(self, rmax, profile_args=None):
        """
        Computes the mass enclosed inside a sphere of radius rmax
        :param rmax: the maximum radius
        :return: mass enclosed inside rmax
        """
        if rmax == 'r200':
            rmax = self.nfw_params[2]
        _integrand = lambda r: 4 * np.pi * r ** 2 * self.density_profile_3d(r, profile_args)
        return quad(_integrand, 0, rmax)[0]



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC as TNFWLenstronomy
import numpy as np


class _TNFWCBaseClass(Halo):

    """
    The base class for a cored and truncated NFW halo
    """
    _scale_evolution_timescale = 1.0
    # the parametric model by Yang et al. (2023) doesn't conserve mass; if this is set to True,
    # then the density profile of each lens is rescaled such that the mass inside the CDM halo equivalent virial radius
    # is conserved.
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._concentration_class = concentration_class
        self._truncation_class = truncation_class
        self._tnfwc_lenstronomy = TNFWLenstronomy()
        mdef = 'TNFWC'
        super(_TNFWCBaseClass, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                           lens_cosmo_instance, args, unique_tag)

    def density_profile_3d(self, r, profile_args=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        if profile_args is None:
            [rho_s_kpc, rs_kpc, rc_kpc, rt_kpc, _] = self.profile_args()
        else:
            [rho_s_kpc, rs_kpc, rc_kpc, rt_kpc, _] = profile_args
        tau = rt_kpc / rs_kpc
        beta = rc_kpc / rs_kpc
        x = r / rs_kpc
        return rho_s_kpc * tau ** 2 / (beta ** 2 + x**2) ** 0.5 / (tau ** 2 + x ** 2) / (1 + x ** 2)

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        return ['TNFWC']

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

class TNFWCFieldHaloSIDM(_TNFWCBaseClass):

    """
    Implements a temporal evolution of the halo profile based on Yang et al. (2023)
    https://arxiv.org/pdf/2305.16176.pdf
    """
    _regularize_t_over_tc = 1.0 # don't evolve profile past this time because fitting function breaks down
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        # This class must have a cross section model supplied through the kwargs_special argument
        self._sidm_cross_section = args['SIDM_CROSS_SECTION']
        _check_valid_cross_section(self._sidm_cross_section)
        super(TNFWCFieldHaloSIDM, self).__init__(mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag)

    @property
    def sidm_timescale(self):
        """
        Computes the timescale given by Equation 2.2 in https://arxiv.org/pdf/2305.16176.pdf
        :return:
        """
        if not hasattr(self, '_sidm_timescale'):
            rho_s, rs, _ = self.nfw_params
            C = 0.75
            G = 4.3e-6
            conversion_fac = 2.135788e-10
            vmax = 1.65 * np.sqrt(G * rho_s * rs ** 2) # fixed for an NFW profile
            v_scale = np.sqrt(4 * np.pi * G * rho_s * rs ** 2)
            sigma_eff = self._sidm_cross_section.effective_cross_section(0.64 * vmax)
            denom = sigma_eff * rho_s * v_scale * conversion_fac
            self._sidm_timescale = self._scale_evolution_timescale * 150 / C / denom
        return self._sidm_timescale

    @property
    def t_over_tc(self):
        """
        Computes the dimensionless timescale for the halo evolution
        :return:
        """
        return self.halo_age / self.sidm_timescale

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_kwargs_lenstronomy'):

            arcsec = 2 * np.pi / 360 / 3600
            [rhos_kpc, rs_kpc, rc_kpc, rt_kpc, _] = self.profile_args()
            dd_kpc = self.lens_cosmo.cosmo.D_A_z(self.z) * 1e3
            Rs_angle, theta_Rs = self._lens_cosmo.nfw_physical2angle_fromNFWparams(1e9*rhos_kpc,
                                                                                   1e-3*rs_kpc,
                                                                                   self.z,
                                                                                   pseudo_nfw=True)
            r_trunc_angle = rt_kpc / dd_kpc / arcsec
            r_core_angle = max(1e-6 * Rs_angle, rc_kpc / dd_kpc / arcsec)
            x, y = np.round(self.x, 4), np.round(self.y, 4)
            kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle,
                       'center_x': x, 'center_y': y, 'r_trunc': r_trunc_angle, 'r_core': r_core_angle}]
            self._kwargs_lenstronomy = kwargs

        return self._kwargs_lenstronomy, None

    def profile_args(self):
        """
        TODO: the parameterization for the density profile presented by Yang et al. doesn't conserve mass.
        I temporarily fix this by rescaling the normalization to conserve mass at all times

        Computes the time-evolving profile parameters
        :return: the density normalization, scale radius, and core radius for thee SIDM halo
        """
        if not hasattr(self, '_profile_args'):
            t = min(self._regularize_t_over_tc, self.t_over_tc)
            rhos_0, rs_0, r200_0 = self.nfw_params
            rho_s = rhos_0 * rho_s_evolution(t)
            rs_kpc = rs_0 * rs_evolution(t)
            rc_kpc = rs_0 * rc_evolution(t)
            rt_kpc = self._truncation_class.truncation_radius_halo(self)
            self._profile_args = (1.0 * rho_s, rs_kpc, rc_kpc, rt_kpc, r200_0)
        return self._profile_args

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_params_physical'):
            [rho_s, rs_kpc, rc_kpc, rt_kpc, r200_kpc] = self.profile_args
            self._params_physical = {'rhos': rho_s * self._rescale_norm,
                                     'rs': rs_kpc, 'r200': r200_kpc,
                                     'r_trunc_kpc': rt_kpc,
                                     'r_core_kcp': rc_kpc}
        return self._params_physical

class TNFWCSubhaloSIDM(TNFWCFieldHaloSIDM):

    """
    Implements a temporal evolution of the halo profile based on Yang et al. (2023)
    https://arxiv.org/pdf/2305.16176.pdf

    TODO: add something here to account for tidal effects
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag):

        if 'scale_evolution_timescale' in args.keys():
            self._scale_evolution_timescale = args['scale_evolution_timescale']
        else:
            self._scale_evolution_timescale = 1.0

        super(TNFWCSubhaloSIDM, self).__init__(mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag)

def _check_valid_cross_section(cross_section):
    """

    :param cross_section:
    :return:
    """
    if getattr(cross_section, 'effective_cross_section') is None:
        raise Exception(
            'the specified SIDM_CROSS_SECTION class must have a method called effective_cross_section that takes'
            'as input a velocity and returns a cross section amplitude in cm^2 / gram')
    if not callable(cross_section.effective_cross_section):
        raise Exception('the effective_cross_section method in the cross section class must be a callable function that '
                        'takes an input a velocity and returns a cross section amplitude in cm^2 / gram.')

def rho_s_evolution(tr):
    """
    Computes the evolution of the density normalization for the profile
    :param t_over_tc: the physical time scaled by the characteristic evolution time
    :return: the density in units of rho_s_nfw
    """

    return 0.692896 + 1.11215 * tr + 4.08729 * tr ** 5 - 5.78038 * tr ** 7 + 5.30498 * tr ** 9 + (
        1 - 0.692896) / np.log(0.001) * np.log(0.001 + tr)

def rs_evolution(tr):
    """
    Computes the evolution of the scale radius for the profile
    :param t_over_tc: the physical time scaled by the characteristic evolution time
    :return: the scale radius in units of rs_nfw
    """

    return 1.13293 - 0.518186 * tr + 0.30265 * tr ** 2 - 0.400463 * tr ** 3 + (1 - 1.13293) * np.log(
        0.001 + tr) / np.log(0.001)

def rc_evolution(tr):
    """
    Computes the evolution of the core radius for the profile
    :param t_over_tc: the physical time scaled by the characteristic evolution time
    :return: the core radius in units of rs_nfw
    """

    return 1.06419 * np.sqrt(tr) - 1.33207 * tr + 0.431133 * tr ** 2 - 0.148087 * tr ** 3 + 0.024455 * tr ** 4



from pyHalo.Halos.galacticus_util.galacticus_util import GalacticusUtil
import numpy as np

def nodedata_apply_filter(nodedata,nodefilter,galacticus_util = None):
    """Takes a dictionary with numpy arrays as values and apply as boolean filter to all.
     Returns a dictionary with same keys, but a filter applied to all arrays."""
    return {key:val[nodefilter] for key,val in nodedata.items()}

def nodedata_select_subhalo(nodedata,id, galacticus_util = None):
    """Selects a subhalo specified by an id, returns a dictionary of the subhalo's properties"""
    gutil = GalacticusUtil() if galacticus_util is None else galacticus_util
    idfilter = nodedata[nodedata[gutil.PARAM_NODE_ID] == id]
    return {key:val[idfilter][0] for key,val in nodedata.items()}

def nodedata_filter_tree(nodedata, treeindex,galacticus_util = None):
    """Returns a filter that excludes all nodes but nodes in the specified tree"""
    gutil = GalacticusUtil() if galacticus_util is None else galacticus_util
    return nodedata[gutil.PARAM_TREE_INDEX] == treeindex

def nodedata_filter_subhalos(nodedata,galacticus_util= None):
    """Returns a filter that excludes all but subhalos (excludes host halos)"""
    gutil = GalacticusUtil() if galacticus_util is None else galacticus_util
    return nodedata[gutil.PARAM_ISOLATED] == 0

def nodedata_filter_halos(nodedata,galacticus_util = None):
    """Returns a filter that excludes all but halo (excludes sub-halos)"""
    return np.logical_not(nodedata_filter_subhalos(nodedata))

def nodedata_filter_range(nodedata,range,key,galacticus_util=None):
    """Returns a filter that excludes nodes not within the given range for a specified parameter"""
    return (nodedata[key] > range[0]) & (nodedata[key] < range[1]) 

def nodedata_filter_virialized(nodedata,galacticus_util= None):
    """
    Returns a filter that excludes everything outside of the host halos virial radius
    WARNING: Current implementation only works if there is only one host halo per tree,
    IE we are looking at the last output from galacticus. 
    """
    gutil = GalacticusUtil() if galacticus_util is None else galacticus_util

    #Get radial position of halos
    rvec = np.asarray((nodedata[gutil.PARAM_X],nodedata[gutil.PARAM_Y],nodedata[gutil.PARAM_Z]))
    r = np.linalg.norm(rvec,axis=0)

    #Filter halos and get there virial radii
    filtered_halos = nodedata_filter_halos(nodedata)
    rv_halos = nodedata[gutil.PARAM_RADIUS_VIRIAL][filtered_halos]
    halo_output_n = nodedata[gutil.PARAM_TREE_ORDER]

    filter_virialized = r < rv_halos[halo_output_n]

    return filter_virialized

def nodedata_filter_r2d(nodedata,r2d_max,plane_normal,
                        galacticus_util= None):
    
    """
    Filters based on projected radii.
    """
    gutil = GalacticusUtil() if galacticus_util is None else galacticus_util
    
    r = np.asarray((nodedata[gutil.PARAM_X],nodedata[gutil.PARAM_Y],nodedata[gutil.PARAM_Z]))

    r2d = project_r2d(*r,plane_normal)

    return r2d < r2d_max

def project_r2d(x,y,z,plane_normal):
    """
    Takes in arrays of coordinates, calculates the projected radius on the plane. 
     

    :param x: An array of x coordinates
    :param y: An array of y coordinates
    :param z: An array of z coordinates
    :plane_normal: Normal vector of the plane to project onto
    """
    coords = np.asarray((x,y,z))

    #Reshape so if scalers are passed for x,y,z we do not encounter issue
    if coords.ndim == 1:
        coords.reshape((3,1))

    #convert to unit normal
    un = plane_normal / np.linalg.norm(plane_normal)

    #Project distance to plane
    #Projected distance to plane is sqrt(r.r - (r.un)^2)
    rdotr = np.linalg.norm(coords,axis=0)**2
    rdotun = np.dot(un,coords)
    return np.sqrt(rdotr - rdotun**2)
    




    



    




import numpy.testing as npt
import pytest
from pyHalo.PresetModels.cdm import CDM
from pyHalo.plotting_routines import *

class TestPlottingRoutines(object):

    def setup_method(self):
        self.realization = CDM(0.5, 1.5, sigma_sub=0.025, LOS_normalization=0.0)

    def test_spatial_distribution(self):
        plot_subhalo_spatial_distribution(self.realization)

    def test_mass_function_plot(self):

        plot_subhalo_mass_functon(self.realization)
        plot_subhalo_mass_functon(self.realization, bound_mass_function=True)
        plot_halo_mass_function(self.realization, z_eval=0.5)
        plot_halo_mass_function(self.realization, z_eval=None)
        plot_halo_mass_function(self.realization, z_eval=[0.5, 0.6])

    def test_bound_mass_function_plot(self):

        plot_subhalo_bound_mass(self.realization)
        plot_bound_mass_histogram(self.realization)
        plot_subhalo_concentration_versus_bound_mass(self.realization)
        plot_subhalo_infall_time_versus_bound_mass(self.realization)

    def test_plot_mc_relation(self):

        plot_concentration_mass_relation(self.realization, z_eval='z_lens')
        plot_concentration_mass_relation(self.realization, z_eval=0.5)
        plot_concentration_mass_relation(self.realization, z_eval=[0.4, 0.6])

    def test_truncation_radius_plot(self):

        plot_truncation_radius_histogram(self.realization, subhalos_only=False)
        plot_truncation_radius_histogram(self.realization, subhalos_only=True)

    def test_convergence_plot(self):

        fig = plt.figure(1)
        ax = plt.subplot(111)
        plot_multiplane_convergence(self.realization, ax=ax, npix=10, show_critical_curve=False)
        plot_multiplane_convergence(self.realization, npix=10, show_critical_curve=False, subtract_mean_kappa=True)
        plot_multiplane_convergence(self.realization, npix=10, show_critical_curve=True)

if __name__ == '__main__':
     pytest.main()



import matplotlib.pyplot as plt
import numpy as np
from pyHalo.single_realization import realization_at_z
from lenstronomy.Util.analysis_util import azimuthalAverage
from pyHalo.utilities import multiplane_convergence

plt.rcParams['axes.linewidth'] = 2.5
plt.rcParams['xtick.major.width'] = 3.5
plt.rcParams['xtick.major.size'] = 10
plt.rcParams['xtick.minor.size'] = 6
plt.rcParams['ytick.minor.size'] = 6
plt.rcParams['ytick.major.width'] = 3.5
plt.rcParams['ytick.major.size'] = 10
plt.rcParams['ytick.labelsize'] = 20
plt.rcParams['xtick.labelsize'] = 20
plt.rcParams['font.family']

def plot_subhalo_spatial_distribution(realization, max_range_arcsec=3.0, ax=None, color='k', kwargs_plot={},
                                     log_mlow=6.0, log_mhigh=10.0, nbins=25):
    """
    Plots the azimuthally-averaged radial distribution of subhalos
    :param realization: an instance of Realization that contains subhalos
    :param ax: figure axis
    :param color:
    :param kwargs_plot:
    :param log_mlow:
    :param log_mhigh:
    :param nbins:
    :return:
    """

    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)
    x_arcsec, y_arcsec = [], []
    for halo in realization.halos:
        if halo.is_subhalo:
            if halo.mass >= 10**log_mlow and halo.mass < 10**log_mhigh:
                x_arcsec.append(halo.x)
                y_arcsec.append(halo.y)
    root2 = np.sqrt(2)
    rangex, rangey = (-max_range_arcsec/root2, max_range_arcsec/root2), (-max_range_arcsec/root2, max_range_arcsec/root2)
    h, binx, biny = np.histogram2d(x_arcsec, y_arcsec, bins=nbins, range=(rangex, rangey))
    radial_prof, r_bin = azimuthalAverage(h)
    arcsec_per_pixel = max_range_arcsec / len(r_bin)
    r_bin_arcsec = r_bin * arcsec_per_pixel
    ax.plot(r_bin_arcsec, radial_prof, color=color, **kwargs_plot)
    ax.set_xlabel(r'$r_{\rm{2D}} \ \left[\rm{arcsec}\right]$', fontsize=16)
    ax.set_ylabel('number density '+r'$\left[\rm{arcsec}^{-2}\right]$', fontsize=16)

def plot_halo_mass_function(realization, z_eval=None, ax=None, color='k', kwargs_plot={},
                                     log_mlow=6.0, log_mhigh=10.0, nbins=25):

    """
    Makes a log-log plot of the halo mass function for a given realization.

    :param realization:
    :param z_eval:
    :param ax:
    :param color:
    :param kwargs_plot:
    :param log_mlow:
    :param log_mhigh:
    :param nbins:
    :param show_errorbars:
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    xlabel = 'halo mass '+r'$\left[M_{\odot}\right]$'
    ylabel = r'$n\left(m\right)$'
    if z_eval is None:
        masses = realization.masses
    elif isinstance(z_eval, float) or isinstance(z_eval, int):
        real = realization_at_z(realization, z_eval)[0]
        masses = real.masses
    elif isinstance(z_eval, list):
        masses = []
        for halo in realization.halos:
            if halo.is_subhalo:
                continue
            if halo.z >= z_eval[0] and halo.z < z_eval[1]:
                masses.append(halo.mass)
    else:
        raise Exception('z_eval must be one of: a specific redshift (i.e. a floating point number), or an'
                        'interval z_eval = [z_min, z_max]')
    h, m = np.histogram(masses, bins=np.logspace(log_mlow, log_mhigh, nbins))
    ax.plot(m[0:-1], h, color=color, **kwargs_plot)
    ax.set_xlabel(xlabel, fontsize=14)
    ax.set_ylabel(ylabel, fontsize=14)
    ax.set_xscale('log')
    ax.set_yscale('log')

def plot_concentration_mass_relation(realization, z_eval, ax=None, color='k', kwargs_plot={},
                                     log_mlow=6.0, log_mhigh=10.0, nbins=20, show_errorbars=False):

    """
    Makes a log-log plot of the concentration-mass relation for a given realization.

    :param realization: an instance of the Realization class
    :param z_eval: the redshift at which to evaluate the mc relation; can be one of 'z_lens', a flaot,
    or an interval [z_min, z_max]
    :param ax: figure axis
    :param color: the color for the lines in the figure
    :param kwargs_plot: keyword arguments passed to plot()
    :param log_mlow: minimum halo mass
    :param log_mhigh: maximum halo mass
    :param nbins: number of mass bins along the x axis
    :param show_errorbars: bool; show the logarithmic scatter
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    if z_eval == 'z_lens':
        xlabel = 'infall halo mass ' + r'$\left[M_{\odot}\right]$'
        ylabel = r'$c\left(m\right)$'
        halo_mass_list = []
        halo_concentration_list = []
        for halo in realization.halos:
            if halo.is_subhalo:
                halo_mass_list.append(halo.mass)
                halo_concentration_list.append(halo.c)
    elif isinstance(z_eval, float) or isinstance(z_eval, int):
        xlabel = 'halo mass ' + r'$\left[M_{\odot}\right]$'
        ylabel = r'$c\left(m\right)$'
        halo_mass_list = []
        halo_concentration_list = []
        for halo in realization.halos:
            if halo.is_subhalo and halo.z == z_eval:
                halo_mass_list.append(halo.mass)
                halo_concentration_list.append(halo.c)
    elif isinstance(z_eval, list):
        xlabel = 'halo mass ' + r'$\left[M_{\odot}\right]$'
        ylabel = r'$c\left(m\right)$'
        halo_mass_list = []
        halo_concentration_list = []
        for halo in realization.halos:
            if halo.z >= z_eval[0] and halo.z < z_eval[1]:
                halo_mass_list.append(halo.mass)
                halo_concentration_list.append(halo.c)
    else:
        raise Exception('z_eval must be one of: z_lens, a specific redshift (i.e. a floating point number), or an'
                        'interval z_eval = [z_min, z_max]')

    halo_masses = np.array(halo_mass_list)
    halo_concentrations = np.array(halo_concentration_list)

    x = np.logspace(log_mlow, log_mhigh, nbins + 1)
    c = []
    c_error = []
    for i in range(0, len(x) - 1):
        cond1 = halo_masses >= x[i]
        cond2 = halo_masses < x[i + 1]
        inds = np.where(np.logical_and(cond1, cond2))[0]
        c_median = np.median(halo_concentrations[inds])
        c_std = np.std(halo_concentrations[inds])
        c.append(c_median)
        c_error.append(c_std)

    if show_errorbars:
        ax.errorbar(x[0:-1], c, yerr=c_error, fmt='none', color=color)
    ax.plot(x[0:-1], c, color=color, **kwargs_plot)
    ax.set_xlabel(xlabel, fontsize=15)
    ax.set_ylabel(ylabel, fontsize=15)
    ax.set_xscale('log')
    ax.set_yscale('linear')

def plot_subhalo_bound_mass(realization, ax=None, color='k', kwargs_plot={},
                            log_mlow=6.0, log_mhigh=10.0):
    """
    Plots the bound mass of subhalos defined as the mass inside the infall virial radius of a truncated
    profile. Note: the halo mass definition must have a truncation radius specified for this method

    :param realization: an instance of Realization
    :param ax: figure axis
    :param color: color for the lines
    :param kwargs_plot: keyword arguments passed to plot
    :param log_mlow: minimum infall halo mass along x axis
    :param log_mhigh: maximum infall halo mass along x axis
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)
    masses = []
    mass_bound_fraction = []
    for halo in realization.halos:
        if halo.is_subhalo:
            masses.append(halo.mass)
            mass_bound_fraction.append(halo.bound_mass)

    ax.scatter(masses, mass_bound_fraction, color=color, **kwargs_plot)
    ax.plot(masses, masses, color='0.5', lw=2, linestyle=':')
    ax.set_xlim(10**log_mlow, 10**log_mhigh)
    ax.set_xlabel('infall halo mass '+r'$\left[M_{\odot}\right]$', fontsize=15)
    ax.set_ylabel('subhalo mass '+r'$M\left(r < r_{200}\right)$', fontsize=15)
    ax.set_xscale('log')
    ax.set_yscale('log')


def plot_subhalo_mass_functon(realization, log_m_low=6.0, log_m_high=10.0, bound_mass_function=False, nbins=20,
                              n_bootstrap=10, ax=None, color='k', kwargs_plot={}, rescale_amp=1.0):
    """
    Makes a log-log plot of the infall mass versus the final bound mass of tidally stripped subhalos
    :param realization: an instance of Realization
    :param log_mlow: minimum infall halo mass along x axis
    :param log_mhigh: maximum infall halo mass along x axis
    :param bound_mass_function:
    :param nbins: number of bins along x axis
    :param n_bootstrap: number of bootstrap resampling iteratins to compute vertical error bars
    :param ax: figure axis
    :param color: color for the lines
    :param kwargs_plot: keyword arguments passed to plot
    :param rescale_amp: rescales the plotted mass function
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)
    masses = []
    for halo in realization.halos:
        if halo.is_subhalo:
            if bound_mass_function:
                masses.append(halo.bound_mass)
            else:
                masses.append(halo.mass)
    h = np.empty((n_bootstrap, nbins))
    for i in range(0, n_bootstrap):
        inds = np.random.randint(0, len(masses), len(masses))
        _h, _x = np.histogram(np.log10(masses)[inds], bins=nbins, range=(log_m_low, log_m_high))
        h[i, :] = _h
    median = np.median(h, axis=0)
    standard_dev = np.std(h, axis=0)
    logm = _x[1:] - (_x[1] - _x[0]) / 2
    ax.errorbar(logm, rescale_amp * median, yerr=standard_dev, fmt='none', color=color)
    ax.plot(logm, rescale_amp * median, color=color, **kwargs_plot)
    ax.set_ylim(median[-1] / 2, median[0] * 2)
    ax.set_xlabel('infall halo mass ' + r'$\left[\log_{10} M_{\odot}\right]$', fontsize=15)
    ax.set_ylabel(r'$n\left(m\right)$', fontsize=15)
    ax.set_yscale('log')
    return (logm, median, standard_dev)

def plot_subhalo_concentration_versus_bound_mass(realization, ax=None, color='k', kwargs_plot={},
                                                 log_mlow=6.0, log_mhigh=10.0):
    """
    Plots the bound mass of subhalos, defined as the mass inside the infall virial radius of a truncated
    profile, versus the infall concentration of the halo.
    Note: the halo mass definition must have a truncation radius specified for this method

    :param realization: an instance of Realization
    :param ax: figure axis
    :param color: color for the lines
    :param kwargs_plot: keyword arguments passed to plot
    :param log_mlow: minimum infall halo mass along x axis
    :param log_mhigh: maximum infall halo mass along x axis
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    infall_concentration = []
    bound_mass_fraction = []
    for halo in realization.halos:
        if halo.is_subhalo:
            if halo.mass >= 10 ** log_mlow and halo.mass < 10 ** log_mhigh:
                infall_concentration.append(halo.c)
                bound_mass_fraction.append(halo.bound_mass / halo.mass)
    ax.scatter(infall_concentration, bound_mass_fraction, color=color, **kwargs_plot)
    ax.set_xlabel('infall concentration', fontsize=15)
    ax.set_ylabel(r'$\log_{10} \frac{M_{\rm{bound}}}{M_{\rm{infall}}}$', fontsize=22)
    # ax.set_xscale('log')
    ax.set_yscale('log')

def plot_subhalo_infall_time_versus_bound_mass(realization, ax=None, color='k', kwargs_plot={},
                                                 log_mlow=6.0, log_mhigh=10.0):
    """
    Plots the time since infall of subhalos versus the infall bound mass of the subhalo.
    Note: the halo mass definition must have a truncation radius specified for this method

    :param realization: an instance of Realization
    :param ax: figure axis
    :param color: color for the lines
    :param kwargs_plot: keyword arguments passed to plot
    :param log_mlow: minimum infall halo mass along x axis
    :param log_mhigh: maximum infall halo mass along x axis
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    infall_time = []
    bound_mass_fraction = []
    for halo in realization.halos:
        if halo.is_subhalo:
            if halo.mass >= 10 ** log_mlow and halo.mass < 10 ** log_mhigh:
                infall_time.append(halo.time_since_infall)
                bound_mass_fraction.append(halo.bound_mass / halo.mass)
    ax.scatter(infall_time, bound_mass_fraction, color=color, **kwargs_plot)
    ax.set_xlabel('time since infall [Gyr]', fontsize=15)
    ax.set_ylabel(r'$\log_{10} \frac{M_{\rm{bound}}}{M_{\rm{infall}}}$', fontsize=22)
    # ax.set_xscale('log')
    ax.set_yscale('log')

def plot_bound_mass_histogram(realization, ax=None, color='k', kwargs_plot={},
                              log_mlow=6.0, log_mhigh=10.0):
    """
    Plots the bound mass of subhalos defined as the mass inside the infall virial radius of a truncated
    profile.

    Note: the halo mass definition must have a truncation radius specified

    :param realization: an instance of Realization
    :param ax: figure axis
    :param color: color for the lines
    :param kwargs_plot: keyword arguments passed to plot
    :param log_mlow: minimum infall halo mass
    :param log_mhigh: maximum infall halo mass
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    bound_mass_fraction = []
    for halo in realization.halos:
        if halo.is_subhalo:
            if halo.mass >= 10 ** log_mlow and halo.mass < 10 ** log_mhigh:
                bound_mass_fraction.append(halo.bound_mass / halo.mass)
    ax.hist(np.log10(bound_mass_fraction), color=color, density=True, **kwargs_plot)
    ax.set_xlabel(r'$\log_{10} \frac{M_{\rm{bound}}}{M_{\rm{infall}}}$', fontsize=22)

def plot_truncation_radius_histogram(realization, subhalos_only=True, ax=None, color='k', kwargs_plot={},
                                   log_mlow=6.0, log_mhigh=10.0):
    """
    Plots a histogram of the truncation radius of a halo divided by the scale radius.

    Note: the halo mass definition must have a truncation radius specified for this method

    :param realization: an instance of Realization
    :param subhalos_only: bool; show only subhalos
    :param ax: figure axis
    :param color: color for the lines
    :param kwargs_plot: keyword arguments passed to plot
    :param log_mlow: minimum infall halo mass
    :param log_mhigh: maximum infall halo mass
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    tau = []
    for halo in realization.halos:
        if subhalos_only and halo.is_subhalo is False: continue
        if halo.mass >= 10 ** log_mlow and halo.mass < 10 ** log_mhigh:
            params = halo.params_physical
            rt_over_rs = params['r_trunc_kpc'] / params['rs']
            tau.append(rt_over_rs)
    ax.hist(np.log10(tau), color=color, density=True, **kwargs_plot)
    ax.set_xlabel(r'$\frac{r_t}{r_s}$', fontsize=15)


def plot_multiplane_convergence(realization, ax=None, npix=100, cone_opening_angle_arcsec=2.5, lens_model_list_macro=None,
                                kwargs_lens_macro=None, redshift_list_macro=None, cmap='bwr', vmin_max=0.05,
                                subtract_mean_kappa=False, show_critical_curve=True, grid_scale_crit_curve=0.025):
    """

    :param realization: an instance of Realization
    :param ax: a matplotlib axes
    :param npix: number of pixels per axis
    :param cone_opening_angle_arcsec: the size of each axis in arcsec
    :param lens_model_list_macro: a list of lens models that desceibe the main deflector
    :param kwargs_lens_macro: keyword arguments for the macro lens models
    :param redshift_list_macro: a list of redshifts for the macro lens models
    :param cmap: the name of a matploblib colormap
    :param vmin_max: color scale normalization for the convergence map
    :param subtract_mean_kappa: bool; subtract the average convergence in the area for visualization purposes
    :param show_critical_curve: bool; plots the critical curve
    :return: matplotlib axis instance
    """

    if ax is None:
        fig = plt.figure()
        fig.set_size_inches(6, 6)
        ax = plt.subplot(111)

    delta_kappa, lens_model, kwargs_lens, _, _ = multiplane_convergence(realization, cone_opening_angle_arcsec, npix, lens_model_list_macro,
                                         kwargs_lens_macro, redshift_list_macro)
    if subtract_mean_kappa:
        delta_kappa -= np.mean(delta_kappa)
    ax.imshow(delta_kappa, extent=[-cone_opening_angle_arcsec/2, cone_opening_angle_arcsec/2, -cone_opening_angle_arcsec/2, cone_opening_angle_arcsec/2],
              cmap=cmap, vmin=-vmin_max, vmax=vmin_max, origin='lower')

    if show_critical_curve:
        from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
        ext = LensModelExtensions(lens_model)
        ra_crit_list, dec_crit_list, ra_caustic_list, dec_caustic_list = ext.critical_curve_caustics(
            kwargs_lens, grid_scale=grid_scale_crit_curve, compute_window=cone_opening_angle_arcsec)
        ax.plot(ra_crit_list[0], dec_crit_list[0], color='k')

    return ax




import numpy as numpy
from colossus.halo.concentration import concentration, peaks
import warnings
import numpy
from scipy.interpolate import RegularGridInterpolator

warnings.filterwarnings("ignore")

__all__ = ['ConcentrationDiemerJoyce',
           'ConcentrationWDMHyperbolic',
           'ConcentrationWDMPolynomial',
           'ConcentrationPeakHeight']

class _ConcentrationCDM(object):

    def __init__(self, cosmo, scatter=True, scatter_dex=0.2, *args, **kwargs):
        """
        This class handles concentrations of the mass-concentration relation for NFW profiles
        :param lens_cosmo: an instance of LensCosmo
        """
        self._cosmo = cosmo
        self._scatter = scatter
        self._scatter_dex = scatter_dex

    def nfw_concentration(self, m, z):
        """
        Evaluates the concentration of a halo of mass 'm' at redshift z
        :param M: halo mass [M_sun]
        :param z: halo redshift
        :return:
        """
        if isinstance(m, float) or isinstance(m, int):
            c = self._evaluate_concentration(m, z)
        else:
            c = numpy.array([self._evaluate_concentration(mi, z) for mi in m])
        if self._scatter:
            log_c = numpy.log(c)
            c = numpy.random.lognormal(log_c, self._scatter_dex)
        return c

    def _evaluate_concentration(self, *args, **kargs):
        raise Exception(
            'Custom concentration class must have a method evaluate_concentration with inputs mass, redshift')

class _ConcentrationTurnover(object):

    def __init__(self, cdm_concentration):
        """

        :param cdm_concentration: an instantiated CDM concentration-mass relation class
        """
        self._cdm_concentration = cdm_concentration

    def nfw_concentration(self, m, z):
        """
        Evaluates the concentration of a halo of mass 'm' at redshift z
        :param M: halo mass [M_sun]
        :param z: halo redshift
        :return:
        """
        c_cdm = self._cdm_concentration.nfw_concentration(m, z)
        return c_cdm * self.suppression(m, z)

    def suppression(self, *args, **kwargs):
        raise Exception('a WDM model with a turnover must have a suppression function')

class ConcentrationDiemerJoyce(_ConcentrationCDM):

    name = 'DIEMERJOYCE19'

    def __init__(self, cosmo, scatter=True, scatter_dex=0.2, mdef='200c', *args, **kwargs):
        """
        This class handles concentrations of the mass-concentration relation for NFW profiles
        :param cosmo: an instance of astropy cosmology
        """
        self._mdef = mdef
        super(ConcentrationDiemerJoyce, self).__init__(cosmo, scatter, scatter_dex)

    def _evaluate_concentration(self, M, z):

        """
        Evaluates the concentration of an NFW profile

        :param M: halo mass; m200 with respect to critical density of the Universe at redshift z
        :param z: redshift
        :return: halo concentratioon
        """
        model = 'diemer19'
        if isinstance(M, float) or isinstance(M, int):
            M_h = M * self._cosmo.h
            c = concentration(M_h, mdef=self._mdef, model=model, z=z)
        else:
            if isinstance(z, numpy.ndarray) or isinstance(z, list):
                assert len(z) == len(M)
                c = []
                for (mi, zi) in zip(M, z):
                    M_h = mi * self._cosmo.h
                    c.append(concentration(M_h, mdef=self._mdef, model=model, z=zi))
            else:
                M_h = M * self._cosmo.h
                c = concentration(M_h, mdef=self._mdef, model=model, z=z)
        return c

class ConcentrationLudlow(_ConcentrationCDM):

    name = 'LUDLOW2016'

    def __init__(self, cosmo, scatter=True, scatter_dex=0.2, mdef='200c', *args, **kwargs):
        """
        This class handles concentrations of the mass-concentration relation for NFW profiles
        :param cosmo: an instance of astropy cosmology
        """
        self._cosmo = cosmo
        self._mdef = mdef
        super(ConcentrationLudlow, self).__init__(cosmo, scatter, scatter_dex)

    # def _setup_model(self):
    #     """
    #     As discussed in the Colossus documentation the evaluation of this model is efficient for large arrays of
    #     halo masses but inefficient for individual halo masses. Therefore, in this routine we will compute the m-c
    #     relation on a grid (efficient) and interpolate it to later sample individual conentrations
    #     :return:
    #     """
    #     n = 40
    #     log10m_array = numpy.log10(numpy.logspace(5, 11, n))
    #     z_array = numpy.linspace(0, 15, n)
    #     values = numpy.empty((n, n))
    #     for i in range(0, len(z_array)):
    #         values[:, i] = self.evaluate_concentration_colossus(10**log10m_array, z_array[i])
    #     points = (log10m_array, z_array)
    #     interp = RegularGridInterpolator(points, values)
    #     return interp

    # def evaluate_concentration_colossus(self, m, z):
    #     """
    #     This function makes a direct call to Colossus to do the calculation of the median halo concentration.
    #     This is inefficient for individual halo masses
    #     :param m: halo mass in units 200c
    #     :param z: halo redshift
    #     :return: halo concentration
    #     """
    #     model = 'ludlow16'
    #     M_h = m * self._cosmo.h
    #     c = concentration(M_h, mdef=self._mdef, model=model, z=z)
    #     return c

    def _evaluate_concentration(self, M, z):

        """
        Evaluates the concentration of an NFW profile

        :param M: halo mass; m200 with respect to critical density of the Universe at redshift z
        :param z: redshift
        :return: halo concentratioon
        """
        # try:
        #     M_h = M * self._cosmo.h
        #     point = (numpy.log10(M_h), z)
        #     c = float(self._interp(point))
        # except:
        #     model = 'ludlow16'
        M_h = M * self._cosmo.h
        z = min(15.0, z)
        c = concentration(M_h, mdef=self._mdef, model='ludlow16', z=z)
        return c

class ConcentrationPeakHeight(_ConcentrationCDM):

    name = 'PEAK_HEIGHT_POWERLAW'

    def __init__(self, cosmo, c0, zeta, beta, scatter=True, scatter_dex=0.2):
        """
        This class handles concentrations of the mass-concentration relation for NFW profiles
        :param cosmo: an instance of astropy cosmology
        :param c0: the amplitude of the concentration-mass relation at 10^8 M_sun at z=0
        :param zeta: modifies the logarithmic slope of the concentration-mass relation (1+z)^zeta
        :param beta: the logarithmic slope of the concentration-mass relation in peak height
        :param scatter: bool; whether to include scatter in concentration-mass relation
        :param scatter_dex: scatter in concentration in dex
        """
        self._c0 = c0
        self._zeta = zeta
        self._beta = beta
        self._redshift_evolution = _zEvolutionPeakHeight(cosmo)
        super(ConcentrationPeakHeight, self).__init__(cosmo, scatter, scatter_dex)

    def _evaluate_concentration(self, M, z):

        """
        Evaluates the concentration of an NFW profile

        :param M: halo mass; m200 with respect to critical density of the Universe at redshift z
        :param z: redshift
        :return: halo concentratioon
        """
        M_h = M * self._cosmo.h
        Mref_h = 10 ** 8 * self._cosmo.h
        nu = peaks.peakHeight(M_h, z)
        nu_ref = peaks.peakHeight(Mref_h, z)
        redshift_factor = self._redshift_evolution(M, z)
        c = self._c0 * (nu / nu_ref) ** -self._beta * redshift_factor ** self._zeta
        return c

class ConcentrationWDMPolynomial(_ConcentrationTurnover):

    name = 'WDM_POLYNOMIAL'

    def __init__(self, cosmo, concentration_cdm_class, log_mc, c_scale=60.0,
                 c_power=-0.17, c_power_inner=1.0, mc_suppression_redshift_evolution=True, scatter=True,
                 scatter_dex=0.2, kwargs_cdm={}):
        """

        :param cosmo: an instance of astropy cosmology
        :param concentration_cdm_class: a concentration class for CDM
        :param c_scale: the leading coefficient of the suppression term (see below)
        :param c_power: the exponent outside the parenthesis of the suppression term (see equation below)
        :param c_power_inner: the exponent inside the parenthesis of the suppression term (see equation below)
        :param mc_suppression_redshift_evolution: bool; adds the (mild) redshift evolution from Bose et al. (2016)
        :param scatter: bool; whether to include scatter in concentration-mass relation
        :param scatter_dex: scatter in concentration in dex
        :param kwargs_cdm: keyword arguments for the CDM concentration class
        """
        if 'scatter' not in kwargs_cdm.keys():
            kwargs_cdm['scatter'] = scatter
        if 'scatter_dex' not in kwargs_cdm.keys():
            kwargs_cdm['scatter_dex'] = scatter_dex
        cdm_concentration = concentration_cdm_class(cosmo, **kwargs_cdm)
        if c_power > 0:
            raise Exception('c_power parameters > 0 are unphysical')
        if c_scale < 0:
            raise Exception('c_scale parameters < 0 are unphysical')
        self._log_mc = log_mc
        self._c_scale = c_scale
        self._c_power = c_power
        self._c_power_inner = c_power_inner
        self._mc_suppression_redshift_evolution = mc_suppression_redshift_evolution
        self._redshift_evolution = _zEvolutionBose2016()
        super(ConcentrationWDMPolynomial, self).__init__(cdm_concentration)

    def suppression(self, m, z):
        """

        :param m: halo mass [same units as log_mc]
        :param z: halo redshift
        :param log_mc: mass scale where suppresion kicks in [same units as m]
        :return:
        """
        mhm = 10 ** self._log_mc
        rescale_factor = (1 + self._c_scale * (mhm / m) ** self._c_power_inner) ** self._c_power
        if self._mc_suppression_redshift_evolution:
            redshift_factor = self._redshift_evolution(m, z)
        else:
            redshift_factor = 1.0
        rescale = redshift_factor * rescale_factor
        return rescale

class ConcentrationWDMHyperbolic(_ConcentrationTurnover):

    name = 'WDM_HYPERBOLIC'

    def __init__(self, cosmo, concentration_cdm_class, log_mc, a,  b, scatter=True,
                 scatter_dex=0.2, kwargs_cdm={}):
        """

        :param cosmo:
        :param concentration_cdm_class:
        :param kwargs_cdm:
        """
        if 'scatter' not in kwargs_cdm.keys():
            kwargs_cdm['scatter'] = scatter
        if 'scatter_dex' not in kwargs_cdm.keys():
            kwargs_cdm['scatter_dex'] = scatter_dex
        cdm_concentration = concentration_cdm_class(cosmo, **kwargs_cdm)
        self._a = a
        self._b = b
        self._log_mc = log_mc
        super(ConcentrationWDMHyperbolic, self).__init__(cdm_concentration)

    def suppression(self, m, z):
        """

        :param m:
        :param z:
        :param log_mc:
        :param a:
        :param b:
        :return:
        """
        mhm = 10 ** self._log_mc
        log10u = numpy.log10(m / mhm)
        argument = (log10u - self._a) / (2 * self._b)
        return 0.5 * (1 + numpy.tanh(argument))

class ConcentrationLudlowWDM(_ConcentrationTurnover):

    name = 'LUDLOW_WDM'

    points = (numpy.array([1.0, 2.0, 2.7, 3.2, 4.0]),
              numpy.array([6.0, 7.0, 8.0]),
              numpy.array([0.0, 1.0, 2.0, 3.0, 4.0]))
    values_a = numpy.array([ 0.554,  0.484,  0.432,  0.276,  0.116,  0.474,  0.46 ,  0.376,
        0.206,  0.053,  0.486,  0.429,  0.319,  0.12 , -0.048,  0.497,
        0.457,  0.37 ,  0.288,  0.131,  0.472,  0.446,  0.35 ,  0.216,
        0.057,  0.477,  0.393,  0.329,  0.134, -0.021,  0.49 ,  0.412,
        0.367,  0.268,  0.121,  0.454,  0.406,  0.343,  0.24 ,  0.074,
        0.455,  0.388,  0.287,  0.123, -0.017,  0.453,  0.428,  0.363,
        0.271,  0.128,  0.442,  0.412,  0.337,  0.225,  0.072,  0.409,
        0.392,  0.282,  0.146,  0.004,  0.432,  0.414,  0.348,  0.255,
        0.103,  0.425,  0.419,  0.318,  0.166,  0.065,  0.412,  0.379,
        0.283,  0.119,  0.01 ])
    values_b = numpy.array([0.805, 0.766, 0.85 , 0.927, 0.938, 0.739, 0.782, 0.868, 0.952,
       1.015, 0.782, 0.778, 0.892, 0.946, 1.045, 0.545, 0.513, 0.552,
       0.617, 0.715, 0.533, 0.574, 0.606, 0.711, 0.721, 0.549, 0.56 ,
       0.638, 0.702, 0.749, 0.499, 0.46 , 0.495, 0.543, 0.646, 0.468,
       0.485, 0.553, 0.57 , 0.686, 0.498, 0.509, 0.56 , 0.64 , 0.684,
       0.457, 0.464, 0.499, 0.492, 0.631, 0.445, 0.471, 0.499, 0.555,
       0.657, 0.45 , 0.504, 0.518, 0.641, 0.685, 0.42 , 0.447, 0.477,
       0.483, 0.584, 0.42 , 0.476, 0.501, 0.565, 0.622, 0.424, 0.473,
       0.514, 0.59 , 0.665])
    interp_a = RegularGridInterpolator(points, values_a.reshape(5, 3, 5))
    interp_b = RegularGridInterpolator(points, values_b.reshape(5, 3, 5))

    def __init__(self, cosmo, log_mc, dlogT_dlogk, scatter=True, scatter_dex=0.2, mdef='200c'):
        """

        :param cosmo:
        :param log_mc:
        :param dlogT_dlogk:
        :param scatter:
        :param scatter_dex:
        :param mdef:
        """
        if dlogT_dlogk > 0:
            raise Exception('positive logarithmic derivatives are unphysical')
        self._dlogT_dlogk = dlogT_dlogk
        self._log_mc = log_mc
        cdm_concentration = ConcentrationLudlow(cosmo, scatter, scatter_dex, mdef)
        super(ConcentrationLudlowWDM, self).__init__(cdm_concentration)

    @staticmethod
    def _make_in_bounds(log10_mhm, dlogT_dlogk, z):
        """
        Forces the values of log10_mhm, dlogT_dlogk, and z to be inside the range of interpolation
        :return: the values of these parameters
        """
        log10_mhm_eval = max(6.0, log10_mhm)
        log10_mhm_eval = min(8.0, log10_mhm_eval)
        dlogT_dlogk_eval = -1.0 * dlogT_dlogk
        dlogT_dlogk_eval = max(1.0, dlogT_dlogk_eval)
        dlogT_dlogk_eval = min(4.0, dlogT_dlogk_eval)
        z_eval = max(z, 0.0)
        z_eval = min(z_eval, 4.0)
        return log10_mhm_eval, dlogT_dlogk_eval, z_eval

    def suppression_fit(self, log10_mhm, dlogT_dlogk, z):
        """
        Evaluates the coefficients of the hyperbolic suppression term
        :param log10_mhm: log10 half-mode mass
        :param dlogT_dlogk: logarithmic derivative at the half-mode scale k_1/2
        :param z: redshift
        :return: suppression of the WDM relation relative to CDM
        """
        # we will only evaluate this model around the scales where it was calibrated; i.e. no extrapolation
        log10_mhm_eval, dlogT_dlogk_eval, z_eval = self._make_in_bounds(log10_mhm, dlogT_dlogk, z)
        x = (dlogT_dlogk_eval, log10_mhm_eval, z_eval)
        a = self.interp_a(x)
        b = self.interp_b(x)
        return numpy.squeeze(a), numpy.squeeze(b)

    def nfw_concentration(self, m, z):
        """
        Evaluates the concentration of a halo of mass 'm' at redshift z
        :param M: halo mass [M_sun]
        :param z: halo redshift
        :return: halo concentration
        """
        c_cdm = self._cdm_concentration.nfw_concentration(m, z)
        return c_cdm * self.suppression(m, z)

    def suppression(self, m, z):
        """
        Evaluates the suppression of the concentration mass relation such that c_wdm = c_cdm * suppression
        :param m: halo mass [solar mass]
        :param z: redshift
        :return: the suppression factor of the WDM concentration-mass relation
        """
        mhm = 10 ** self._log_mc
        a, b = self.suppression_fit(self._log_mc, self._dlogT_dlogk, z)
        a = numpy.squeeze(a)
        b = numpy.squeeze(b)
        log10u = numpy.log10(m / mhm)
        arg = (log10u - a) / (2 * b)
        return 0.5 * (1 + numpy.tanh(arg))

class _zEvolutionPeakHeight(object):

    def __init__(self, cosmo):
        self._cosmo = cosmo

    def __call__(self, m, z):
        """
        This method evaluates the redshift evolution according to the redshift evolution of the peak height
        :param m: halo mass in units 200c
        :param z: redshift
        :return: the relative evolution of the peak height between z=0 and z=z
        """
        M_h = m * self._cosmo.h
        redshift_factor = peaks.peakHeight(M_h, 0.0) / peaks.peakHeight(M_h, z)
        return redshift_factor

class _zEvolutionBose2016(object):

    def __call__(self, m, z):
        """
        This method evaluates the redshift evolution according to the redshift evolution for WDM halos presented by
        Bose 2016
        :param m: halo mass in units 200c
        :param z: redshift
        :return: the scaling with redshift of the Bose et al. (2016) model
        """
        redshift_factor = (1 + z) ** (0.026 * z - 0.04)
        return redshift_factor



from pyHalo.PresetModels.cdm import CDM, CDMCorrelatedStructure
from pyHalo.PresetModels.wdm import WDM, WDM_mixed
from pyHalo.PresetModels.sidm import SIDM_core_collapse
from pyHalo.PresetModels.uldm import ULDM
from pyHalo.preset_models import preset_model_from_name
from pyHalo.PresetModels.external import CDMFromEmulator, DMFromGalacticus
from pyHalo.Halos.galacticus_util.galacticus_util import GalacticusUtil
from pyHalo.Halos.HaloModels.TNFWFromParams import TNFWFromParams
import pytest
import numpy as np
import numpy.testing as npt
from copy import copy
from pyHalo.Cosmology.cosmology import Cosmology


class TestPresetModels(object):

    def test_CDM(self):

        cdm = CDM(0.5, 1.5)
        _ = cdm.lensing_quantities()
        _ = preset_model_from_name('CDM')

        cdm2 = CDM(0.6, 1.5, log_m_host=13.3,
                  truncation_model_subhalos='TRUNCATION_GALACTICUS')
        cdm3 = CDM(0.6, 1.5, log_m_host=13.3,
                   truncation_model_subhalos='TRUNCATION_GALACTICUS',
                   host_scaling_factor=4.0, redshift_scaling_factor=2.0)
        halos2 = len(cdm2.halos)
        halos3 = len(cdm3.halos)
        npt.assert_equal(halos3 > halos2, True)

    def test_CDM_correlated_structure_only(self):

        cdm = CDMCorrelatedStructure(0.5, 1.5)
        _ = cdm.lensing_quantities()

    def test_WDM(self):

        wdm = WDM(0.5, 1.5, 8.0)
        _ = wdm.lensing_quantities()
        _ = preset_model_from_name('WDM')

        _ = WDM(0.5, 1.5, 8.0,
                  truncation_model_subhalos='TRUNCATION_GALACTICUS')

    def test_ULDM(self):

        flucs_shape = 'ring'
        flucs_args = {'angle': 0.0, 'rmin': 0.9, 'rmax': 1.1}
        uldm = ULDM(0.5, 1.5, -21, flucs_shape=flucs_shape, flucs_args=flucs_args)
        _ = uldm.lensing_quantities()
        _ = preset_model_from_name('ULDM')

    def test_SIDM_core_collapse(self):
        mass_ranges_subhalos = [[6, 8], [8, 10]]
        mass_ranges_field_halos = [[6, 8], [8, 10]]
        probabilities_subhalos = [1, 1]
        probabilities_field_halos = [1, 1]
        sidm_cc = SIDM_core_collapse(0.5, 1.5, mass_ranges_subhalos, mass_ranges_field_halos,
        probabilities_subhalos, probabilities_field_halos)
        _ = sidm_cc.lensing_quantities()
        _ = preset_model_from_name('SIDM_core_collapse')

    def test_WDM_mixed(self):
        wdm_mixed = WDM_mixed(0.5, 1.5, 8.0, 0.5)
        _ = wdm_mixed.lensing_quantities()
        _ = preset_model_from_name('WDM_mixed')

    def test_WDM_general(self):
        func = preset_model_from_name('WDMGeneral')
        wdm = func(0.5, 1.5, 7.7, -2.0)
        _ = wdm.lensing_quantities()
        wdm = func(0.5, 1.5, 7.7, -2.0, truncation_model_subhalos='TRUNCATION_GALACTICUS')

    def test_CDM_emulator(self):

        def emulator_input_callable(*args, **kwargs):
            subhalo_infall_masses = np.array([10**7,10**8])
            subhalo_x_kpc = np.array([1.0, 1.0])
            subhalo_y_kpc = np.array([1.0, 1.0])
            subhalo_final_bound_masses = subhalo_infall_masses / 2
            subhalo_infall_concentrations = np.array([16.0, 20.0])
            return subhalo_infall_masses, subhalo_x_kpc, subhalo_y_kpc, subhalo_final_bound_masses, subhalo_infall_concentrations

        concentrations = np.array([16.0, 20.0])
        mass_array = np.array([10 ** 7, 10 ** 8])
        kwargs_cdm = {'LOS_normalization': 0.0}
        cdm_subhalo_emulator = CDMFromEmulator(0.5, 1.5, emulator_input_callable, kwargs_cdm)
        _ = cdm_subhalo_emulator.lensing_quantities()
        for i, halo in enumerate(cdm_subhalo_emulator.halos):
            npt.assert_equal(halo.mass, mass_array[i])
            npt.assert_almost_equal(halo.x, 0.1584666, 4)
            npt.assert_almost_equal(halo.y, 0.1584666, 4)
            npt.assert_equal(halo.c, concentrations[i])

        emulator_input_array = np.empty((2, 5))
        emulator_input_array[:, 0] = mass_array
        emulator_input_array[:, 1] = np.array([1.0, 1.0])
        emulator_input_array[:, 2] = np.array([1.0, 1.0])
        emulator_input_array[:, 3] = mass_array / 2
        emulator_input_array[:, 4] = concentrations
        cdm_subhalo_emulator = CDMFromEmulator(0.5, 1.5, emulator_input_array, kwargs_cdm)
        _ = cdm_subhalo_emulator.lensing_quantities()
        for i, halo in enumerate(cdm_subhalo_emulator.halos):
            npt.assert_equal(halo.mass, mass_array[i])
            npt.assert_almost_equal(halo.x, 0.1584666, 4)
            npt.assert_almost_equal(halo.y, 0.1584666, 4)
            npt.assert_equal(halo.c, concentrations[i])

    def test_galacticus(self):
        util = GalacticusUtil()

        cosmo = Cosmology()

        # Simulates data loaded from a galacticus hdf5 file.
        # mock_data is in NOT intended to be physical
        # It is however designed to test aspects of DMfromGalacticus
        mock_data = {
            util.PARAM_X:                       np.asarray((1,1,0,1,0  ,  1,0  ,   0,1,0  )),
            util.PARAM_Y:                       np.asarray((1,0,1,0,0  ,  1,0  ,   3,0,0  )),
            util.PARAM_Z:                       np.asarray((0,1,1,0,0  ,  0,0  ,   0,0,0  )),
            util.PARAM_TNFW_RHO_S:              np.asarray((1,1,1,1,1  ,  1,1  ,   1,1,1  )),
            util.PARAM_TNFW_RADIUS_TRUNCATION:  np.asarray((1,1,1,1,1  ,  1,1  ,   1,1,1  )),
            util.PARAM_RADIUS_VIRIAL:           np.asarray((2,2,2,2,2  ,  2,2  ,   2,2,2  )),
            util.PARAM_RADIUS_SCALE:            np.asarray((1,1,1,1,2  ,  1,2  ,   2,2,2  )),
            util.PARAM_MASS_BOUND:              np.asarray((1,2,3,4,2  ,  1,2  ,   1,1,2  )),
            util.PARAM_MASS_INFALL:             np.asarray((5,6,7,1,2  ,  1,2  ,   1,1,2  )),
            util.PARAM_ISOLATED:                np.asarray((0,0,0,0,1  ,  0,1  ,   0,0,1  )),
            util.PARAM_TREE_ORDER:              np.asarray((0,0,0,0,0  ,  1,1  ,   2,2,2  )),
            util.PARAM_TREE_INDEX:              np.asarray((1,1,1,1,1  ,  2,2  ,   3,3,3  )),
            util.PARAM_NODE_ID:                 np.asarray((0,1,2,3,4  ,  5,6  ,   8,9,7  )),
            util.PARAM_Z_LAST_ISOLATED:         np.asarray((1,2,3,2,0.5,  1,0.5,   3,1,0.5))
        }

        mock_data[util.PARAM_CONCENTRATION] = mock_data[util.PARAM_RADIUS_VIRIAL] / mock_data[util.PARAM_RADIUS_SCALE]

        kwargs_base = dict(
            galacticus_hdf5=                        mock_data,
            z_source=                               2,
            cone_opening_angle_arcsec=              1E10,
            tree_index=                             1,
            log_mlow_galacticus=                    -10,
            log_mhigh_galacticus=                   10,
            mass_range_is_bound=                    False,
            proj_angle_theta=                       np.pi/2,
            proj_angle_phi=                         0,
            nodedata_filter=                        None,
            galacticus_utilities=                   util,
            galacticus_params_additional=           None,
            galacticus_tabulate_tnfw_params=        None,
            preset_model_los=                       "CDM",
            LOS_normalization=                      0.0
        )

        MPC_TO_KPC = 1E3
        MPC_TO_AS = MPC_TO_KPC / cosmo.kpc_proper_per_asec(0.5)

        kwargs_test_projection = copy(kwargs_base)
        realization_test_projection = DMFromGalacticus(**kwargs_test_projection)
        assert len(realization_test_projection.halos) == 4
        npt.assert_almost_equal(np.linalg.norm(np.asarray((realization_test_projection.halos[0].x,realization_test_projection.halos[0].y))), MPC_TO_AS * 1)
        npt.assert_almost_equal(np.linalg.norm(np.asarray((realization_test_projection.halos[1].x,realization_test_projection.halos[1].y))), MPC_TO_AS * 1)
        npt.assert_almost_equal(np.linalg.norm(np.asarray((realization_test_projection.halos[2].x,realization_test_projection.halos[2].y))), MPC_TO_AS * np.sqrt(2))
        npt.assert_almost_equal(np.linalg.norm(np.asarray((realization_test_projection.halos[3].x,realization_test_projection.halos[3].y))), 0)

        kwargs_test_filter_mass_basic = copy(kwargs_base)
        kwargs_test_filter_mass_basic.update(dict(log_mlow_galacticus=np.log10(4.99),log_mhigh_galacticus=np.log10(6.01)
                                                  ,mass_range_is_bound=False))
        realization_test_filter_mass_basic = DMFromGalacticus(**kwargs_test_filter_mass_basic)
        npt.assert_equal(len(realization_test_filter_mass_basic.halos),2)

        kwargs_test_filter_mass_bound = copy(kwargs_base)
        kwargs_test_filter_mass_bound.update(dict(log_mlow_galacticus=np.log10(0.99),log_mhigh_galacticus=np.log10(2.01)
                                                ,mass_range_is_bound=True))
        realization_test_filter_mass_bound = DMFromGalacticus(**kwargs_test_filter_mass_bound)
        npt.assert_equal(len(realization_test_filter_mass_bound.halos),2)

        kwargs_test_volume_exclusion = copy(kwargs_base)
        kwargs_test_volume_exclusion["cone_opening_angle_arcsec"] = 0.99 * MPC_TO_AS * 2
        realization_test_volume_exclusion = DMFromGalacticus(**kwargs_test_volume_exclusion)
        npt.assert_equal(len(realization_test_volume_exclusion.halos),1)

        kwargs_test_tree2 = copy(kwargs_base)
        kwargs_test_tree2["tree_index"] = 2
        realization_test_tree2 = DMFromGalacticus(**kwargs_test_tree2)
        npt.assert_equal(len(realization_test_tree2.halos),1)

        kwargs_test_exclude_beyond_virial = copy(kwargs_base)
        kwargs_test_exclude_beyond_virial["tree_index"] = 3
        realization_test_exclude_beyond_virial = DMFromGalacticus(**kwargs_test_exclude_beyond_virial)
        npt.assert_equal(len(realization_test_exclude_beyond_virial.halos),1)

        kwargs_test_params = copy(kwargs_base)
        kwargs_test_params["nodedata_filter"] = lambda nd,u: np.ones(nd[u.PARAM_MASS_INFALL].shape[0],dtype=bool)
        realization_test_params = DMFromGalacticus(**kwargs_test_params)
        for n,sh in enumerate(realization_test_params.halos):
            npt.assert_almost_equal(sh.params_physical[TNFWFromParams.KEY_RHO_S],mock_data[util.PARAM_TNFW_RHO_S][n] * 4 * 1 / MPC_TO_KPC**3)
            npt.assert_almost_equal(sh.params_physical[TNFWFromParams.KEY_RS],mock_data[util.PARAM_RADIUS_SCALE][n] * MPC_TO_KPC)
            npt.assert_almost_equal(sh.params_physical[TNFWFromParams.KEY_RV],mock_data[util.PARAM_RADIUS_VIRIAL][n] * MPC_TO_KPC)
            npt.assert_almost_equal(sh.params_physical[TNFWFromParams.KEY_RT],mock_data[util.PARAM_TNFW_RADIUS_TRUNCATION][n] * MPC_TO_KPC)
            npt.assert_almost_equal(sh.z,0.5)
            npt.assert_almost_equal(sh.z_infall,mock_data[util.PARAM_Z_LAST_ISOLATED][n])

if __name__ == '__main__':
    pytest.main()



import numpy as np
from scipy.spatial.transform import Rotation
from pyHalo.Halos.HaloModels.TNFWemulator import TNFWSubhaloEmulator
from pyHalo.Halos.HaloModels.TNFWFromParams import TNFWFromParams
from pyHalo.PresetModels.cdm import CDM
from pyHalo.single_realization import Realization
from pyHalo.Halos.galacticus_util.galacticus_util import GalacticusUtil
from pyHalo.Halos.galacticus_util.galacticus_filter import nodedata_filter_subhalos,nodedata_filter_tree,nodedata_filter_virialized,nodedata_filter_range,nodedata_apply_filter
import h5py


def DMFromGalacticus(galacticus_hdf5,z_source,cone_opening_angle_arcsec,tree_index,log_mlow_galacticus,log_mhigh_galacticus,
                     mass_range_is_bound=True, proj_angle_theta=0, proj_angle_phi=0,
                     nodedata_filter=None, galacticus_utilities=None, galacticus_params_additional=None,
                     galacticus_tabulate_tnfw_params=None, preset_model_los="CDM", **kwargs_los):
    """
    This generates a realization of halos using subhalo parameters provided from a specified tree in the galacticus file.
    See https://github.com/galacticusorg/galacticus/ for information on the galacticus galaxy formation model.

    :param galacticus_hdf5: Galacticus output hdf5 file. If str loads file from specified path
    :param z_source: source redshift
    :param cone_opening_angle_arcsec: The opening angle of the cone in arc seconds
    :param tree_index: The number of the tree to create a realization from.
        A single galacticus file contains multiple realizations (trees).
        NOTE: Trees output from galacticus are assigned a number starting at 1.
    :param log_galacticus_mlow: The log_10 of the minimum mass subhalo to include in the realization.
        Subhalos are filtered by bound or infall mass as set by setting mass_range_is_bound.
    :param log_galacticus_mhigh: The log_10 of the minimum subhalo to include in the realization
        Subhalos are filtered by bound or infall mass as set by setting mass_range_is_bound.
    :param mass_range_is_bound: If true subhalos are filtered bound mass, if false subhalos are filtered by infall mass.
    :param proj_angle_theta: Specifies the theta angle (in spherical coordinates) of the normal vector for the plane to project subhalo coordinates on.
        Should be in the range [0,pi]
    :param proj_angle_theta: Specifies the theta angle (in spherical coordinates) of the normal vector for the plane to project subhalo coordinates on.
        Should be in the range [0, 2 * pi]
    :param nodedata_filter: Expects a callable function that has input and output: (dict[str,np.ndarray], GalacticusUtil) -> np.ndarray[bool]
        ,subhalos are filtered based on the output np array. Defaults to None. If provided overrides default filtering.
    :param galacticus_params: Extra parameters to read when loading in galacticus hdf5 file.
    :param galacticus_tabulate_params: Expects a callable function that has input and output: (dict[str,np.ndarray], GalacticusUtil) -> dict[str,np.ndarray]
        Should return a dictionary containing a numpy array of args for use in creation of tnfw halos.
    :param preset_model_los: Specifies the preset model to use when generating the LOS subhalos. Defaults to CDM

    :return: A realization from Galacticus halos

    NOTE:
    For galacticus output files: All trees should output at the same redshift. The final output should include only one host halo per tree.
    An example galacticus output and it's parameter file can be found in example_notebooks/data
    """
    #Avoid circular import
    from pyHalo.preset_models import preset_model_from_name

    gutil = GalacticusUtil() if galacticus_utilities is None else galacticus_utilities

    MPC_TO_KPC = 1E3

    #Only read needed parameters to save memory.
    PARAMS_TO_READ_DEF = [
                            gutil.PARAM_X,
                            gutil.PARAM_Y,
                            gutil.PARAM_Z,
                            gutil.PARAM_TNFW_RHO_S,
                            gutil.PARAM_TNFW_RADIUS_TRUNCATION,
                            gutil.PARAM_RADIUS_VIRIAL,
                            gutil.PARAM_RADIUS_SCALE,
                            gutil.PARAM_MASS_BOUND,
                            gutil.PARAM_MASS_INFALL,
                            gutil.PARAM_ISOLATED,
                            gutil.PARAM_Z_LAST_ISOLATED,
                            gutil.PARAM_CONCENTRATION
                         ]

    if galacticus_params_additional is None:
        galacticus_params_additional = []
    else:
        galacticus_params_additional = list(galacticus_params_additional)

    params_to_read = galacticus_params_additional + PARAMS_TO_READ_DEF

    if isinstance(galacticus_hdf5,str):
        nodedata = gutil.read_nodedata_galacticus(galacticus_hdf5,params_to_read=params_to_read)
    elif isinstance(galacticus_hdf5, h5py.Group):
        nodedata = gutil.hdf5_read_galacticus_nodedata(galacticus_hdf5,params_to_read=params_to_read)
    else:
        nodedata = galacticus_hdf5

    z_lens = np.mean(nodedata[gutil.PARAM_Z_LAST_ISOLATED][np.logical_not(nodedata_filter_subhalos(nodedata,gutil))])

    tree_index = int(np.round(tree_index))

    # we create a realization of only line-of-sight halos by setting sigma_sub = 0.0
    kwargs_los['sigma_sub'] = 0.0
    kwargs_los["cone_opening_angle_arcsec"] = cone_opening_angle_arcsec
    kwargs_los["z_lens"] = z_lens
    kwargs_los["z_source"] = z_source

    halos_LOS = preset_model_from_name(preset_model_los)(**kwargs_los)

    # get lens_cosmo class from class containing LOS objects; note that this will work even if there are no LOS halos
    lens_cosmo = halos_LOS.lens_cosmo

    theta,phi = proj_angle_theta,proj_angle_phi

    # This rotation rotation maps the coordinates such that in the new coordinates zh = nh
    # Then we apply this rotation to x and y unit vectors so we get the x and y unit vectors in the plane
    rotation = Rotation.from_euler("zyz",(0,theta,phi))

    coords = np.asarray((nodedata[gutil.PARAM_X],nodedata[gutil.PARAM_Y],nodedata[gutil.PARAM_Z])) * MPC_TO_KPC

    xh_r = rotation.apply(np.array((1,0,0)))
    yh_r = rotation.apply(np.array((0,1,0)))

    kpc_per_arcsec_at_z = lens_cosmo.cosmo.kpc_proper_per_asec(z_lens)

    r2dmax_kpc = (cone_opening_angle_arcsec / 2) * kpc_per_arcsec_at_z

    coords_2d = np.asarray((np.dot(xh_r,coords),np.dot(yh_r,coords)))
    r2d_mag = np.linalg.norm(coords_2d,axis=0)

    filter_r2d = r2d_mag < r2dmax_kpc
    mass_key = gutil.PARAM_MASS_BOUND if mass_range_is_bound else gutil.PARAM_MASS_INFALL
    mass_range = 10.0**np.asarray((log_mlow_galacticus,log_mhigh_galacticus))

    # We should exclude nodes that are not valid subhalos, such as host halo nodes and nodes that are outside the virial radius.
    # (Galacticus is not calibrated for nodes outside of virial radius)
    if nodedata_filter is None:
        filter_subhalos = nodedata_filter_subhalos(nodedata,gutil)
        filter_virialized = nodedata_filter_virialized(nodedata,gutil)
        filter_mass = nodedata_filter_range(nodedata,mass_range,mass_key,gutil)
        filter_tree = nodedata_filter_tree(nodedata,tree_index,gutil)
        filter_combined = filter_subhalos & filter_virialized & filter_mass & filter_tree & filter_r2d

    else:
        filter_combined = nodedata_filter(nodedata,gutil)

    nodedata = nodedata_apply_filter(nodedata,filter_combined)
    coords_2d = coords_2d[:,filter_combined]
    r2d_mag = r2d_mag[filter_combined]
    coords = coords[:,filter_combined]
    r3d_mag = np.linalg.norm(coords,axis=0)

    def tabualate_params(nodedata,gutil):
        return {
            # The rhos_s factor of 4 comes from the this galacticus output is
            # The density normalization of the underlying NFW halo at r = rs
            # Multiply by 4 to get the normalization for the halo profile
            TNFWFromParams.KEY_RHO_S:       4 * nodedata[gutil.PARAM_TNFW_RHO_S] / (MPC_TO_KPC)**3,
            TNFWFromParams.KEY_RS :         nodedata[gutil.PARAM_RADIUS_SCALE] * MPC_TO_KPC,
            TNFWFromParams.KEY_RT :         nodedata[gutil.PARAM_TNFW_RADIUS_TRUNCATION] * MPC_TO_KPC,
            TNFWFromParams.KEY_RV :         nodedata[gutil.PARAM_CONCENTRATION] * nodedata[gutil.PARAM_RADIUS_SCALE] * MPC_TO_KPC,
            TNFWFromParams.KEY_Z_INFALL:    nodedata[gutil.PARAM_Z_LAST_ISOLATED],
            TNFWFromParams.KEY_ID :         nodedata[gutil.PARAM_NODE_ID]
        }

    galacticus_tabulate_tnfw_params = tabualate_params if galacticus_tabulate_tnfw_params is None else galacticus_tabulate_tnfw_params
    tnfw_args_all = galacticus_tabulate_tnfw_params(nodedata,gutil)

    halo_list = []
    for n,m_infall in enumerate(nodedata[gutil.PARAM_MASS_INFALL]):
        x,y = coords_2d[0][n], coords_2d[1][n]

        tnfw_args = {key:val[n] for key,val in tnfw_args_all.items()}

        halo_list.append(TNFWFromParams(m_infall,x,y,r3d_mag[n],z_lens,True,lens_cosmo,tnfw_args))

    subhalos_from_params = Realization.from_halos(halo_list,lens_cosmo,kwargs_halo_model={},
                                                    msheet_correction=False, rendering_classes=None)

    return halos_LOS.join(subhalos_from_params)

def CDMFromEmulator(z_lens, z_source, emulator_input, kwargs_cdm):
    """
    This generates a realization of subhalos using an emulator of the semi-analytic modeling code Galacticus, and
     generates line-of-sight halos from a mass function parameterized as Sheth-Tormen.

    :param z_lens: main deflector redshift
    :param z_source: sourcee redshift
    :param emulator_input: either an array or a callable function

    if callable: a function that returns an array of
    1) subhalo masses at infall [M_sun]
    2) subhalo projected x position [kpc]
    3) subhalo projected y position [kpc]
    4) subhalo final_bound_mass [M_sun]
    5) subhalo concentrations at infall

    if not callable: an array with shape (N_subhalos 5) that contains masses, positions x, positions y, etc.

    Mass convention is m200 with respect to the critical density of the Universe at redshift Z_infall, where Z_infall is
    the infall redshift (not necessarily the redshift at the time of lensing).

    :param cone_opening_angle_arcsec: the opening angle of the double cone rendering volume in arcsec
    :param log_mlow: log10(minimum halo mass) rendered, or a function that returns log_mlow given a redshift
    :param log_mhigh: log10(maximum halo mass) rendered, or a function that returns log_mlow given a redshift
    :param LOS_normalization: rescaling of the line of sight halo mass function relative to Sheth-Tormen
    :param log_m_host: log10 host halo mass in M_sun
    :param kwargs_other: allows for additional keyword arguments to be specified when creating realization

    The following optional keywords specify a concentration-mass relation for field halos parameterized as a power law
    in peak height. If they are not set in the function call, pyHalo assumes a default concentration-mass relation from Diemer&Joyce
    :param c0: amplitude of the mass-concentration relation at 10^8
    :param log10c0: logarithmic amplitude of the mass-concentration relation at 10^8 (only if c0_mcrelation is None)
    :param beta: logarithmic slope of the mass-concentration-relation pivoting around 10^8
    :param zeta: modifies the redshift evolution of the mass-concentration-relation
    :param two_halo_contribution: whether to include the two-halo term for correlated structure near the main deflector
    :param kwargs_halo_mass_function: keyword arguments passed to the LensingMassFunction class
    (see Cosmology.lensing_mass_function)
    :return: a realization of CDM halos
    """

    # we create a realization of only line-of-sight halos by setting sigma_sub = 0.0
    kwargs_cdm['sigma_sub'] = 0.0
    cdm_halos_LOS = CDM(z_lens, z_source, **kwargs_cdm)
    # get lens_cosmo class from class containing LOS objects; note that this will work even if there are no LOS halos
    lens_cosmo = cdm_halos_LOS.lens_cosmo

    # now create subhalos from the specified properties using the TNFWSubhaloEmulator class
    halo_list = []
    if callable(emulator_input):
        subhalo_infall_masses, subhalo_x_kpc, subhalo_y_kpc, subhalo_final_bound_masses, \
        subhalo_infall_concentrations = emulator_input()
    else:
        subhalo_infall_masses = emulator_input[:, 0]
        subhalo_x_kpc = emulator_input[:, 1]
        subhalo_y_kpc = emulator_input[:, 2]
        subhalo_final_bound_masses = emulator_input[:, 3]
        subhalo_infall_concentrations = emulator_input[:, 4]

    for i in range(0, len(subhalo_infall_masses)):
        halo = TNFWSubhaloEmulator(subhalo_infall_masses[i],
                                   subhalo_x_kpc[i],
                                   subhalo_y_kpc[i],
                                   subhalo_final_bound_masses[i],
                                   subhalo_infall_concentrations[i],
                                   z_lens, lens_cosmo)
        halo_list.append(halo)

    # combine the subhalos with line-of-sight halos
    subhalos_from_emulator = Realization.from_halos(halo_list, lens_cosmo, kwargs_halo_model={},
                                                    msheet_correction=False, rendering_classes=None)
    return cdm_halos_LOS.join(subhalos_from_emulator)



import pytest
import numpy.testing as npt
from pyHalo.Halos.tidal_truncation import TruncationGalacticus, TruncationGalacticusKeeley24
from pyHalo.Halos.HaloModels.TNFW import TNFWSubhalo
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
import numpy as np
from pyHalo.Halos.lens_cosmo import LensCosmo

class TestTruncationGalacticus(object):

    def setup_method(self):

        self.lens_cosmo_instance = LensCosmo(0.5, 1.5)
        chost = 5.0
        self.tg = TruncationGalacticus(self.lens_cosmo_instance, chost)
        self.tgk24 = TruncationGalacticusKeeley24(self.lens_cosmo_instance, chost)

    def test_interp_k24(self):

        interp = self.tgk24._mass_loss_interp

        log10c = 0.5
        m1 = interp.evaluate_mean_mass_loss(log10c, 1.0, 5.0)
        log10c = 2.0
        m2 = interp.evaluate_mean_mass_loss(log10c, 1.0, 5.0)
        npt.assert_equal(m2 > m1, True)

        m = interp.evaluate_mean_mass_loss(np.array([0.5, 2.0]), 1.0, 5.0)
        npt.assert_equal(m[0], m1)
        npt.assert_equal(m[1], m2)

        log10c = 0.5
        m1 = interp.evaluate_scatter_dex(log10c, 1.0, 5.0)
        log10c = 2.0
        m2 = interp.evaluate_scatter_dex(log10c, 1.0, 5.0)
        npt.assert_equal(m1 > m2, True)

        m = interp.evaluate_scatter_dex(np.array([0.5, 2.0]), 1.0, 5.0)
        npt.assert_equal(m[0], m1)
        npt.assert_equal(m[1], m2)

    def test_evaluate_keeley24(self):

        halo_mass = 10**8
        infall_concentration = 10.0
        time_since_infall = 2.0
        chost = 5.0
        z_eval_angles = 0.5
        rt_kpc = self.tgk24.truncation_radius(halo_mass, infall_concentration, time_since_infall,
                                        chost, z_eval_angles)
        npt.assert_equal(True, np.isfinite(rt_kpc))
        npt.assert_equal(True, isinstance(rt_kpc, float))

        halo_mass = 10 ** 8
        infall_concentration = 10.0
        time_since_infall = 8.0
        chost = 5.0
        z_eval_angles = 0.5
        rt_kpc = self.tgk24.truncation_radius(halo_mass, infall_concentration, time_since_infall,
                                            chost, z_eval_angles)

        npt.assert_equal(True, np.isfinite(rt_kpc))
        npt.assert_equal(True, isinstance(rt_kpc, float))

        class DummyHalo(object):

            def __init__(self):
                self.mass = 10**8
                self.c = 16.0
                self.time_since_infall = 5.0
                self.z = 0.5

        halo = DummyHalo()
        rt_kpc = self.tgk24.truncation_radius_halo(halo)
        npt.assert_equal(True, np.isfinite(rt_kpc))
        npt.assert_equal(True, isinstance(rt_kpc, float))

    def test_evaluate(self):

        halo_mass = 10**8
        infall_concentration = 10.0
        time_since_infall = 2.0
        z_eval_angles = 0.5
        rt_kpc = self.tg.truncation_radius(halo_mass, infall_concentration, time_since_infall,
                                        z_eval_angles)
        npt.assert_equal(True, np.isfinite(rt_kpc))
        npt.assert_equal(True, isinstance(rt_kpc, float))

        halo_mass = 10 ** 8
        infall_concentration = 10.0
        time_since_infall = 8.0
        z_eval_angles = 0.5
        rt_kpc = self.tg.truncation_radius(halo_mass, infall_concentration, time_since_infall,
                                            z_eval_angles)

        npt.assert_equal(True, np.isfinite(rt_kpc))
        npt.assert_equal(True, isinstance(rt_kpc, float))

        class DummyHalo(object):

            def __init__(self):
                self.mass = 10**8
                self.c = 16.0
                self.time_since_infall = 5.0
                self.z = 0.5
            def rescale_normalization(self, rescale):
                return

        halo = DummyHalo()
        rt_kpc = self.tg.truncation_radius_halo(halo)
        npt.assert_equal(True, np.isfinite(rt_kpc))
        npt.assert_equal(True, isinstance(rt_kpc, float))

    def test_tg(self):
        lens_cosmo = LensCosmo(0.5, 2.0,None)
        truncation_class = TruncationGalacticus(lens_cosmo, c_host=4.0)
        concentration_class = ConcentrationDiemerJoyce(lens_cosmo.cosmo, scatter=False)
        mass = 10 ** 8
        x = 0
        y = 0
        r3d = None
        z = 0.5
        sub_flag = True
        halo_args = {}
        unique_tag = 1
        tnfw_profile_1 = TNFWSubhalo(mass, x, y, r3d, z, sub_flag,
                                     lens_cosmo, halo_args, truncation_class, concentration_class, unique_tag)
        tnfw_profile_1._c = 3.0
        tnfw_profile_1._time_since_infall = 6.0

        m_bound = tnfw_profile_1.bound_mass
        rescale_norm = tnfw_profile_1._rescale_norm
        kwargs, _ = tnfw_profile_1.lenstronomy_params
        npt.assert_equal(rescale_norm<1, True)
        npt.assert_equal(m_bound < mass, True)

if __name__ == '__main__':
    pytest.main()



import numpy.testing as npt
import numpy as np
from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC as TNFWCLenstronomy
from lenstronomy.LensModel.Profiles.nfw import NFW as NFWLenstronomy
from pyHalo.Halos.HaloModels.NFW_core_trunc import TNFWCFieldHaloSIDM, TNFWCSubhaloSIDM, \
    rho_s_evolution, rs_evolution, rc_evolution, _check_valid_cross_section
from pyHalo.Halos.HaloModels.NFW import NFWFieldHalo
from pyHalo.truncation_models import ConstantTruncationArcsec
from pyHalo.concentration_models import ConcentrationDiemerJoyce
from pyHalo.Halos.lens_cosmo import LensCosmo
import pytest


class TestTNFWC(object):

    def setup_method(self):

        class DummyCrossSection(object):

            def effective_cross_section(self, v):
                return 200.0

        dummy_cross_section = DummyCrossSection()
        mass = 10 ** 8
        self.lens_cosmo = LensCosmo(0.5, 2.5)
        self.tnfwc_lenstronomy = TNFWCLenstronomy()
        self.nfw_lenstronomy = NFWLenstronomy()
        self.truncation_class = ConstantTruncationArcsec(self.lens_cosmo, 1000.0)
        self.concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo.cosmo, scatter=False)
        kwargs_profile = {'SIDM_CROSS_SECTION': dummy_cross_section}
        self.tnfwc = TNFWCFieldHaloSIDM(mass, 0.0, 0.0, None, 0.5, False,
                                        self.lens_cosmo, kwargs_profile, self.truncation_class,
                                        self.concentration_class, 1.0)

        self.tnfw_subhalo = TNFWCSubhaloSIDM(mass, 0.0, 0.0, None, 0.5, True,
                                        self.lens_cosmo, kwargs_profile, self.truncation_class,
                                        self.concentration_class, 1.0)

    def _get_profile(self, t, scale_evolution_timescale=1.0, is_subhalo=False):

        class DummyCrossSection(object):
            def effective_cross_section(self, v):
                return 200.0

        dummy_cross_section = DummyCrossSection()
        mass = 10 ** 8
        kwargs_profile = {'SIDM_CROSS_SECTION': dummy_cross_section,
                          'scale_evolution_timescale': scale_evolution_timescale,
                          'evaluate_mc_at_zlens': True}
        if is_subhalo:
            profile = TNFWCSubhaloSIDM(mass, 0.0, 0.0, None, 0.5, is_subhalo,
                                         self.lens_cosmo, kwargs_profile, self.truncation_class,
                                         self.concentration_class, 1.0)
        else:
            profile = TNFWCFieldHaloSIDM(mass, 0.0, 0.0, None, 0.5, is_subhalo,
                                        self.lens_cosmo, kwargs_profile, self.truncation_class,
                                        self.concentration_class, 1.0)
        profile._halo_age = t
        return profile

    def _get_nfw_profile(self):

        mass = 10 ** 8
        self.lens_cosmo = LensCosmo(0.5, 2.5)
        kwargs_profile = {}
        profile = NFWFieldHalo(mass, 0.0, 0.0, None, 0.5, False,
                                        self.lens_cosmo, kwargs_profile, self.truncation_class,
                                        self.concentration_class, 1.0)
        return profile

    def test_mass_t0(self):

        profile = self._get_profile(0.0)
        nfw_profile = self._get_nfw_profile()
        mass3d = profile.mass_3d('r200')
        mass3d_nfw = nfw_profile.mass_3d('r200')
        npt.assert_almost_equal(mass3d_nfw/mass3d, 1.0, 4)
        npt.assert_almost_equal(mass3d/10**8, 1.0, 4.0)

    def test_valid_cross_section(self):

        class Invalid1(object):
            pass
        class Invalid2(object):
            @property
            def effective_cross_section(self):
                return 1.0
        class Valid(object):
            def effective_cross_section(self, v):
                return 1.0

        npt.assert_raises(Exception,_check_valid_cross_section, Invalid1)
        npt.assert_raises(Exception, _check_valid_cross_section, Invalid2)
        _check_valid_cross_section(Valid)

    def test_rho_evolution(self):

        t_over_tc = 0.0
        rho = rho_s_evolution(t_over_tc)
        npt.assert_almost_equal(rho, 1.0, 5)

    def test_rs_evolution(self):

        t_over_tc = 0.0
        rs = rs_evolution(t_over_tc)
        npt.assert_almost_equal(rs, 1.0, 5)

    def test_rc_evolution(self):

        t_over_tc = 0.0
        rc = rc_evolution(t_over_tc)
        npt.assert_almost_equal(rc, 0.0, 5)

    def test_profile_evolution(self):

        profile0 = self._get_profile(0.0)
        rhos_0, rs_0, _ = profile0.nfw_params
        t_sidm = profile0.sidm_timescale

        t1 = 0.0 * t_sidm
        t2 = 0.5 * t_sidm
        t3 = 1.0 * t_sidm
        profile1 = self._get_profile(t1)
        profile2 = self._get_profile(t2)
        profile3 = self._get_profile(t3)

        kwargs_1 = profile1.lenstronomy_params[0][0]
        kwargs_2 = profile2.lenstronomy_params[0][0]
        kwargs_3 = profile3.lenstronomy_params[0][0]

        arcsec = 2 * np.pi / 360 / 3600
        dd_kpc = self.lens_cosmo.cosmo.D_A_z(profile0.z) * 1e3
        rs1 = rs_0 * rs_evolution(0.0)
        rs2 = rs_0 * rs_evolution(0.5)
        rs3 = rs_0 * rs_evolution(1.0)
        npt.assert_almost_equal(kwargs_1['Rs'] * dd_kpc * arcsec, rs1)
        npt.assert_almost_equal(kwargs_2['Rs'] * dd_kpc * arcsec, rs2)
        npt.assert_almost_equal(kwargs_3['Rs'] * dd_kpc * arcsec, rs3)

        rc1 = rs_0 * rc_evolution(0.0)
        rc2 = rs_0 * rc_evolution(0.5)
        rc3 = rs_0 * rc_evolution(1.0)
        npt.assert_almost_equal(kwargs_1['r_core'] * dd_kpc * arcsec, rc1, 5)
        npt.assert_almost_equal(kwargs_2['r_core'] * dd_kpc * arcsec, rc2, 5)
        npt.assert_almost_equal(kwargs_3['r_core'] * dd_kpc * arcsec, rc3, 5)

        rhos1 = rhos_0 * rho_s_evolution(0.0)
        rhos2 = rhos_0 * rho_s_evolution(0.5)
        rhos3 = rhos_0 * rho_s_evolution(1.0)
        (rho_s, rs_kpc, rc_kpc, rt_kpc, r200_0) = profile1.profile_args()
        npt.assert_almost_equal(rhos1, rho_s)
        (rho_s, rs_kpc, rc_kpc, rt_kpc, r200_0) = profile2.profile_args()
        npt.assert_almost_equal(rhos2, rho_s)
        (rho_s, rs_kpc, rc_kpc, rt_kpc, r200_0) = profile3.profile_args()
        npt.assert_almost_equal(rhos3, rho_s)

    def test_mass_evolution(self):

        m = []
        halo_ages = np.linspace(0.0, 10.0, 50)

        for i, ti in enumerate(halo_ages):
            prof = self._get_profile(ti)
            if i==0:
                r200_0 = prof.nfw_params[2]
            m.append(prof.mass_3d(r200_0))

        m = np.array(m)/m[0]
        npt.assert_allclose(m, 1.0, 1)

    def test_timescale(self):
        profile1 = self._get_profile(1.0)
        profile2 = self._get_profile(1.0, 2.0)
        proifle3 = self._get_profile(1.0, 2.0, is_subhalo=True)
        t1, t2 = profile1.t_over_tc, profile2.t_over_tc
        t3 = proifle3.t_over_tc
        npt.assert_almost_equal(t1/t2, 1.0)
        npt.assert_almost_equal(t1/t3, 2.0)


if __name__ == '__main__':
    pytest.main()



import numpy as np
from scipy.integrate import quad
#from pyHalo.defaults import *

class Geometry(object):

    _delta_z_min = 1e-4

    def __init__(self, cosmology, z_lens, z_source, opening_angle, geometry_type,
                 angle_pad=0.95):

        """
        This class handles geometric calculations for rendering halos along the line of sight

        :param cosmology: an instance of the Cosmology() class in pyHalo
        :param z_lens: lens redshift
        :param z_source: source redshift
        :param opening_angle: the opening angle of the rendering area in arcsec.
        For a double cone configuration, for example, the angular size of the area in which halos are rendered
        is equal to 2 * cone_opening_angle for z < z_lens, and closes towards the source redshift
        :param geometry_type: the type of geometry
        DOUBLE_CONE opens towards the lens redshift and closes behind it
        CYLINDER generates objects in a cylinder; the physical radius of the cylinder is the angular diameter distance
        to the lens times 2 * cone_opening_angle (in radians)
        :param angle_pad: modifies the closure of the rendering volume behind the main deflector, only relevant for
        DOUBLE_CONE geometries. Values less than one will result in a non-zero rendering area at the source redshift
        """

        if geometry_type == 'DOUBLE_CONE':
            self._geometrytype = DoubleCone(cosmology, z_lens, z_source, opening_angle, angle_pad)
            self.volume_type = 'DOUBLE_CONE'
        elif geometry_type == 'CYLINDER':
            self._geometrytype = Cylinder(cosmology, z_lens, z_source, opening_angle)
            self.volume_type = 'CYLINDER'
        elif geometry_type == 'CONE':
            self._geometrytype = Cone(cosmology, z_lens, z_source, opening_angle, angle_pad)
        #elif geometry_type == 'DOUBLE_CONE_CYLINDER':
        #    self._geometrytype = DoubleConeCylindner(cosmology, z_lens, z_source, opening_angle, angle_pad)
        else:
            raise Exception('geometry type '+str(geometry_type) + ' not recognized.')

        self.cosmo = cosmology
        self._zlens, self._zsource = z_lens, z_source
        self.cone_opening_angle = opening_angle
        self._arcsec = self.cosmo.arcsec
        self.kpc_per_arcsec_zlens = self.cosmo.kpc_proper_per_asec(self._zlens)
        self._reduced_to_phys = self._geometrytype._reduced_to_phys

    def rendering_scale(self, z):

        return self._geometrytype.rendering_scale(z)

    def kpc_per_arcsec(self, z):

        return self.cosmo.kpc_proper_per_asec(z)

    def angle_to_physicalradius(self, radius_arcsec, z):

        angle_radian = radius_arcsec * self._arcsec
        return angle_radian * self.rendering_scale(z) * self.cosmo.D_A_z(z)

    def angle_to_comovingradius(self, radius_arcsec, z):

        """
        This is specific to the geometry, e.g. if you pick cone if will close behind the main
        deflector
        """
        a_z = self.cosmo.scale_factor(z)
        return self.angle_to_physicalradius(radius_arcsec, z) / a_z

    def volume_element_comoving(self, z, delta_z, radius=None):
        """
        Computes the volume of a disk at redshift z according to the specified geometry
        :param z: redshift
        :param delta_z: thickness of the redshift pancake
        :param radius: angular radius of the rendering area in arcseconds
        :return: volume element in comoving Mpc for small delta_z
        """

        if radius is None:
            # note that the rendering scale is handled in the angle to co-moving radius function
            radius = 0.5*self.cone_opening_angle
        if delta_z > self._delta_z_min:
            func = self._volume_integrand_comoving
            volume_element = quad(func, z, z+delta_z, args=(radius))[0]
        else:
            volume_element = self._volume_integrand_comoving(z, radius) * delta_z

        return volume_element

    def _volume_integrand_comoving(self, z, radius_arcsec):
        """

        :param theta:
        :param z_lens:
        :param z:
        :return: integrand element in comoving Mpc (for use in scipy.integrate quad)
        """
        area_comoving = self.angle_to_comoving_area(radius_arcsec, z)
        dR = self.cosmo.astropy.hubble_distance.value * self.cosmo.astropy.efunc(z) ** -1
        return area_comoving * dR

    def angle_to_comoving_area(self, radius_arcsec, z):
        """
        computes the area corresponding to the angular radius of a plane at redshift z for a double cone with base at z_base
        :param radius_arcsec: lens cone opening angle in arcsec
        :param z: redshift of plane
        :param z_lens: redshift of main lens
        :return: comoving area
        """
        r = self.angle_to_comovingradius(radius_arcsec, z)
        return np.pi * r ** 2

    def angle_to_physical_area(self, radius_arcsec, z):
        """
        computes the area corresponding to the angular radius of a plane at redshift z for a double cone with base at z_base
        :param theta: lens cone opening angle in arcsec
        :param z: redshift of plane
        :param z_lens: redshift of main lens
        :return: comoving area
        """

        area_comoving = self.angle_to_comoving_area(radius_arcsec, z)

        a_z = self.cosmo.scale_factor(z)

        return area_comoving * a_z ** 2

class Cone(object):

    def __init__(self, cosmology, z_lens, z_source, opening_angle, angle_pad):

        self.cosmo = cosmology

        self.opening_angle_radians = opening_angle * cosmology.arcsec

        self.d_c_lens = cosmology.D_C_transverse(z_lens)

        self._reduced_to_phys = self.cosmo.D_A(0, z_source) / self.cosmo.D_A(z_lens, z_source)

        self.comoving_radius_cylinder = 0.5 * self.opening_angle_radians * self.d_c_lens

        self._zlens, self._zsource = z_lens, z_source

        self._total_volume = np.pi * self.comoving_radius_cylinder ** 2 * \
                             cosmology.D_C_transverse(self._zsource)

    def rendering_scale(self, z):
        """
        Determines the angular size of the area at redshift z where halos are rendered;
        for this geometry, the physical size of the cylinder remains constant, and is set by
        cone_opening_angle / 2 * D_z, where D_z is the distance to the main deflector

        :param z: redshift
        :return: a factor that scales the size of the rendering area
        """
        return 1.0

class Cylinder(object):

    def __init__(self, cosmology, z_lens, z_source, opening_angle):

        self.cosmo = cosmology

        self.opening_angle_radians = opening_angle * cosmology.arcsec

        self.d_c_lens = cosmology.D_C_transverse(z_lens)

        self._reduced_to_phys = self.cosmo.D_A(0, z_source) / self.cosmo.D_A(z_lens, z_source)

        self.comoving_radius_cylinder = 0.5 * self.opening_angle_radians * self.d_c_lens

        self._zlens, self._zsource = z_lens, z_source

        self._total_volume = np.pi * self.comoving_radius_cylinder ** 2 * \
                             cosmology.D_C_transverse(self._zsource)

    def rendering_scale(self, z):
        """
        Determines the angular size of the area at redshift z where halos are rendered;
        for this geometry, the physical size of the cylinder remains constant, and is set by
        cone_opening_angle / 2 * D_z, where D_z is the distance to the main deflector

        :param z: redshift
        :return: a factor that scales the size of the rendering area
        """
        d_c = self.cosmo.D_C_transverse(z)
        xi = self.d_c_lens/d_c

        return xi

class DoubleCone(object):

    def __init__(self, cosmology, z_lens, z_source, opening_angle, angle_pad):

        self.cosmo = cosmology
        self._angle_pad = angle_pad
        d_c_lens = self.cosmo.D_C_transverse(z_lens)
        d_c_lens_source = self.cosmo.D_C_transverse(z_source) - d_c_lens
        comoving_radius_zlens = 0.5 * opening_angle * self.cosmo.arcsec * d_c_lens
        self._reduced_to_phys = self.cosmo.D_A(0, z_source) / self.cosmo.D_A(z_lens, z_source)
        self._zlens, self._zsource = z_lens, z_source
        volume_foreground = (np.pi/3) * comoving_radius_zlens ** 2 * d_c_lens
        volume_background = (np.pi/3) * comoving_radius_zlens ** 2 * d_c_lens_source
        self._total_volume = volume_background + volume_foreground

    def rendering_scale(self, z):
        """
        Determines the angular size of the area at redshift z where halos are rendered;
        for this geometry, the angle stays constant for z < z_lens, and closes towards the source
        angle_pad < 1.0 ensures that the rendering area doesn't close exactly at the source, so that halos are
        rendered up to z = z_source
        :param z: redshift
        :return: a factor that scales the size of the rendering area
        """
        if z <= self._zlens:
            return 1.
        else:
            D_dz = self.cosmo.D_A(self._zlens, z)
            D_z = self.cosmo.D_A_z(z)
            ratio = D_dz / D_z

            return 1 - self._angle_pad * self._reduced_to_phys * ratio



import numpy as np
from copy import deepcopy
from pyHalo.Rendering.line_of_sight import LineOfSightNoSheet
from scipy.integrate import quad
from pyHalo.concentration_models import ConcentrationDiemerJoyce


class TwoHaloContribution(LineOfSightNoSheet):

    name = 'TWO_HALO_TERM'
    """
    This class adds correlated structure associated with the host dark matter halo. The amount of structure added is
    proportional to b * corr, where b is the halo bias as computed by Sheth and Tormen (1999) and corr is the
    matter-matter correlation function. Currently, this term is implemented as a rescaling of the background density by
    b * corr, where the product is the average value computed over 2*dz, where dz is the spacing of the redshift planes
    adjacent the redshift plane of the main deflector.
    """

    def __init__(self, mass_function_model, kwargs_mass_function, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list, use_Lazar_correction=True):

        """

        :param mass_function_model:
        :param kwargs_mass_function:
        :param spatial_distribution_model:
        :param geometry:
        :param lens_cosmo:
        :param lens_plane_redshifts:
        :param use_Lazar_correction:
        :param delta_z_list:
        """
        if 'host_m200' in kwargs_mass_function.keys():
            host_m200 = kwargs_mass_function['host_m200']
        elif 'log_m_host' in kwargs_mass_function.keys():
            host_m200 = 10 ** kwargs_mass_function['log_m_host']
        else:
            raise Exception('must specify the host halo mass through keyword argument host_m200 or log_m_host (base 10)'
                            'when adding the two-halo term!')
        concentration_model = ConcentrationDiemerJoyce(lens_cosmo.cosmo.astropy, scatter=False)
        c_host = concentration_model.nfw_concentration(host_m200, lens_cosmo.z_lens)
        _, _, r200_host_kpc = lens_cosmo.NFW_params_physical(host_m200, c_host, lens_cosmo.z_lens)
        r200_host_mpc = r200_host_kpc * 1e-3
        z_eval = lens_cosmo.z_lens
        idx = np.argmin(abs(np.array(lens_plane_redshifts) - z_eval))
        delta_z = delta_z_list[idx]
        boost = two_halo_enhancement_factor(z_eval, delta_z, lens_cosmo, host_m200, r200_host_mpc, use_Lazar_correction)
        kwargs_mass_function_scaled = deepcopy(kwargs_mass_function)
        kwargs_mass_function_scaled['LOS_normalization'] *= boost
        self._delta_z = delta_z
        super(TwoHaloContribution, self).__init__(mass_function_model, kwargs_mass_function_scaled, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list)

    def render(self):
        """
        Generates halo masses and positions for correlated structure near the main deflector.
        These objects are placed at the lens redshift
        :return: mass (in Msun), x (arcsec), y (arcsec), r3d (kpc), redshift, subhalo_flag (bool)
        """
        mfunc_model = self._get_mass_function_model(self._lens_cosmo.z_lens, self._delta_z)
        masses = mfunc_model.draw()
        nhalos = len(masses)
        x, y = self.render_positions_at_z(self._lens_cosmo.z_lens, nhalos)
        redshifts = np.array([self._lens_cosmo.z_lens] * nhalos)
        subhalo_flag = [False] * nhalos
        r3d = np.array([None] * nhalos)
        return masses, x, y, r3d, redshifts, subhalo_flag

def modificationLazar2021(r, r200, b_e=0.1, s_e=4.0):
    """
    Modification to the two-halo contribution around the main deflector using Equation 3 from Lazar et al. (2021)
    https://ui.adsabs.harvard.edu/abs/2021MNRAS.502.6064L/abstract
    :param r: distance in Mpc
    :param r200: host halo virial radius in Mpc
    :param b_e: calibrated parameter
    :param s_e: calibrated parameter
    :return: the correction factor for the background density that gives a better match to simulations
    """
    return b_e * (r / r200 / 5) ** -s_e

def _boost_integrand(r, m200_host, z, r200_host, lens_cosmo, use_Lazar_correction):
    """
    The local enhancement of the background density of the Universe from the two-halo term and a correction factor
    see Lazar et al. 2021
    :param r: distance from the host in Mpc
    :param m200_host: host halo mass
    :param z: host halo redshift
    :param r200_host: host halo virial radius
    :param lens_cosmo: an instance of LensCosmo
    :return: the local enhancement to the background density
    """
    if use_Lazar_correction:
        return lens_cosmo.twohaloterm(r, m200_host, z) + modificationLazar2021(r, r200_host)
    else:
        return lens_cosmo.twohaloterm(r, m200_host, z)

def two_halo_enhancement_factor(z_lens, z_step, lens_cosmo, overdensity_m200, r200_host, use_Lazar_correction):
    """
    Calculates the enhancement of the background density due to correlated structure around the main deflector. Includes
    a contribution from the two-halo term and a correction factor calibrated by Lazar et al. (2021)
    :param z_lens: deflector redshift
    :param z_step: redshift spacing
    :param lens_cosmo: instance of LensCosmo
    :param overdensity_m200: host halo mass
    :param use_Lazar_correction: bool; adds the correction derived by Lazar et al. (2021)
    :return: the enhancement of the background density per unit length
    """
    rmax = lens_cosmo.cosmo.D_C_transverse(z_lens + z_step) - lens_cosmo.cosmo.D_C_transverse(z_lens)
    rmin = min(rmax, 0.5)
    args = (overdensity_m200, z_lens, r200_host, lens_cosmo, use_Lazar_correction)
    two_halo_boost = 2 * quad(_boost_integrand, rmin, rmax, args=args)[0] / (rmax - rmin)
    return two_halo_boost
